TRANSACTION_CREATE_REAL = 1
TRANSACTION_CREATE_FAKE = 2
TRANSACTION_PARSE = 3

INS_NONE = 0x00
INS_RESET = 0x02

INS_GET_KEY = 0x20
INS_PUT_KEY = 0x22
INS_GET_CHACHA8_PREKEY = 0x24
INS_VERIFY_KEY = 0x26
INS_MANAGE_SEEDWORDS = 0x28

INS_SECRET_KEY_TO_PUBLIC_KEY = 0x30
INS_GEN_KEY_DERIVATION = 0x32
INS_DERIVATION_TO_SCALAR = 0x34
INS_DERIVE_PUBLIC_KEY = 0x36
INS_DERIVE_SECRET_KEY = 0x38
INS_GEN_KEY_IMAGE = 0x3A
INS_SECRET_KEY_ADD = 0x3C
INS_SECRET_KEY_SUB = 0x3E
INS_GENERATE_KEYPAIR = 0x40
INS_SECRET_SCAL_MUL_KEY = 0x42
INS_SECRET_SCAL_MUL_BASE = 0x44

INS_DERIVE_SUBADDRESS_PUBLIC_KEY = 0x46
INS_GET_SUBADDRESS = 0x48
INS_GET_SUBADDRESS_SPEND_PUBLIC_KEY = 0x4A
INS_GET_SUBADDRESS_SECRET_KEY = 0x4C

INS_OPEN_TX = 0x70
INS_SET_SIGNATURE_MODE = 0x72
INS_GET_ADDITIONAL_KEY = 0x74
INS_STEALTH = 0x76
INS_BLIND = 0x78
INS_UNBLIND = 0x7A
INS_VALIDATE = 0x7C
INS_MLSAG = 0x7E
INS_CLOSE_TX = 0x80

INS_GET_RESPONSE = 0xC0
IN_OPTION_MASK = 0x000000FF
OUT_OPTION_MASK = 0x0000FF00
IN_OPTION_MORE_COMMAND = 0x00000080


SW_OK = 0x9000
SW_ALGORITHM_UNSUPPORTED = 0x9484
SW_BYTES_REMAINING_00 = 0x6100
SW_WARNING_STATE_UNCHANGED = 0x6200
SW_STATE_TERMINATED = 0x6285
SW_MORE_DATA_AVAILABLE = 0x6310
SW_WRONG_LENGTH = 0x6700
SW_LOGICAL_CHANNEL_NOT_SUPPORTED = 0x6881
SW_SECURE_MESSAGING_NOT_SUPPORTED = 0x6882
SW_LAST_COMMAND_EXPECTED = 0x6883
SW_COMMAND_CHAINING_NOT_SUPPORTED = 0x6884
SW_SECURITY_LOAD_KEY = 0x6900
SW_SECURITY_COMMITMENT_CONTROL = 0x6911
SW_SECURITY_AMOUNT_CHAIN_CONTROL = 0x6912
SW_SECURITY_COMMITMENT_CHAIN_CONTROL = 0x6913
SW_SECURITY_STATUS_NOT_SATISFIED = 0x6982
SW_FILE_INVALID = 0x6983
SW_PIN_BLOCKED = 0x6983
SW_DATA_INVALID = 0x6984
SW_CONDITIONS_NOT_SATISFIED = 0x6985
SW_COMMAND_NOT_ALLOWED = 0x6986
SW_APPLET_SELECT_FAILED = 0x6999
SW_WRONG_DATA = 0x6A80
SW_FUNC_NOT_SUPPORTED = 0x6A81
SW_FILE_NOT_FOUND = 0x6A82
SW_RECORD_NOT_FOUND = 0x6A83
SW_FILE_FULL = 0x6A84
SW_INCORRECT_P1P2 = 0x6A86
SW_REFERENCED_DATA_NOT_FOUND = 0x6A88
SW_WRONG_P1P2 = 0x6B00
SW_CORRECT_LENGTH_00 = 0x6C00
SW_INS_NOT_SUPPORTED = 0x6D00
SW_CLA_NOT_SUPPORTED = 0x6E00
SW_UNKNOWN = 0x6F00



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


from monero_glue.compat import gc, log

from apps.monero.controller import misc
from monero_glue.messages import MessageType


class TsxSigner(object):
    """
    Monero Transaction signer.
    Provides interface to the host, packages messages.
    """

    def __init__(self):
        from apps.monero.controller import iface

        self.ctx = None
        self.err_ctr = 0
        self.tsx_obj = None  # type: TTransactionBuilder
        self.creds = None  # type: apps.monero.xmr.sub.creds.AccountCreds
        self.iface = iface.get_iface()
        self.debug = True
        self.purge = False

    async def tsx_exc_handler(self, e):
        """
        Handles the exception thrown in the Trezor processing. Clears transaction state.
        We could use decorator/wrapper for message calls but not sure how uPython handles them
        so now are entry points wrapped in try-catch.

        :param e:
        :return:
        """
        if self.debug:
            log.warning(__name__, "Transaction exception: %s: %s", type(e), e)

        self.err_ctr += 1
        self.purge = True
        self.tsx_obj = None  # clear transaction object
        await self.iface.transaction_error(e)

    async def should_purge(self):
        """
        Delete global state?
        :return:
        """
        return self.purge or (self.tsx_obj and self.tsx_obj.is_terminal())

    def _log_trace(self, x=None, collect=False):
        log.debug(
            __name__, "Log trace: %s, ... F: %s A: %s", x, gc.mem_free(), gc.mem_alloc()
        )
        if collect:
            gc.collect()

    async def setup(self, msg):
        from apps.monero.controller import wrapper

        self.creds = await wrapper.monero_get_creds(
            self.ctx, msg.address_n or (), msg.network_type
        )

    async def restore(self, state):
        from monero_glue.protocol.tsx_sign_builder import TTransactionBuilder

        self.tsx_obj = TTransactionBuilder(self, creds=self.creds, state=state)
        self._log_trace("Restored", True)

    async def state_save(self):
        try:
            s = self.tsx_obj.state_save()
            self.tsx_obj = None
        finally:
            gc.collect()
        self._log_trace("State saved")
        return s

    async def wake_up(self, ctx, state, msg, iface=None):
        """
        Restore from the stored state
        :param state:
        :return:
        """
        from apps.monero.controller import iface as mod_iface

        self.ctx = ctx
        self.iface = mod_iface.get_iface(ctx) if not iface else iface
        gc.collect()

        self._log_trace("wake_up()", True)

        if msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionInitRequest:
            self._log_trace("init")
            await self.setup(msg)

        await self.restore(
            state
            if msg.MESSAGE_WIRE_TYPE != MessageType.MoneroTransactionInitRequest
            else None
        )
        self._log_trace("wake_up() end", True)

    async def sign(self, msg):
        """
        Main multiplex point
        :param msg:
        :return:
        """
        self._log_trace("sign()", True)
        if msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionInitRequest:
            self._log_trace("sign_init")
            return await self.tsx_init(msg.tsx_data)
        elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionSetInputRequest:
            self._log_trace("sign_inp")
            return await self.tsx_set_input(msg)
        elif (
            msg.MESSAGE_WIRE_TYPE
            == MessageType.MoneroTransactionInputsPermutationRequest
        ):
            self._log_trace("sign_perm")
            return await self.tsx_inputs_permutation(msg)
        elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionInputViniRequest:
            self._log_trace("sign_vin")
            return await self.tsx_input_vini(msg)
        elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionAllInputsSetRequest:
            self._log_trace("all_in_set")
            return await self.tsx_all_in_set(msg)
        elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionSetOutputRequest:
            self._log_trace("sign_out")
            return await self.tsx_set_output1(msg)
        elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionAllOutSetRequest:
            self._log_trace("sign_out_set")
            return await self.tsx_all_out1_set(msg)
        elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionSignInputRequest:
            self._log_trace("sign_sinp")
            return await self.tsx_sign_input(msg)
        elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroTransactionFinalRequest:
            self._log_trace("sign_final")
            return await self.tsx_sign_final(msg)
        else:
            raise ValueError("Unknown message")

    async def tsx_init(self, tsx_data):
        """
        Initialize transaction state.
        :param tsx_data:
        :return:
        """
        try:
            return await self.tsx_obj.init_transaction(tsx_data)
        except Exception as e:
            await self.tsx_exc_handler(e)
            raise

    async def tsx_set_input(self, msg):
        """
        Sets UTXO one by one.
        Computes spending secret key, key image. tx.vin[i] + HMAC, Pedersen commitment on amount.

        If number of inputs is small, in-memory mode is used = alpha, pseudo_outs are kept in the Trezor.
        Otherwise pseudo_outs are offloaded with HMAC, alpha is offloaded encrypted under AES-GCM() with
        key derived for exactly this purpose.

        :param msg
        :return:
        """
        try:
            return await self.tsx_obj.set_input(msg.src_entr)
        except Exception as e:
            await self.tsx_exc_handler(e)
            raise

    async def tsx_inputs_permutation(self, msg):
        """
        Set permutation on the inputs - sorted by key image on host.

        :return:
        """
        try:
            return await self.tsx_obj.tsx_inputs_permutation(msg.perm)
        except Exception as e:
            await self.tsx_exc_handler(e)
            raise

    async def tsx_input_vini(self, msg):
        """
        Set tx.vin[i] for incremental tx prefix hash computation.
        After sorting by key images on host.

        :return:
        """
        try:
            return await self.tsx_obj.input_vini(
                msg.src_entr,
                msg.vini,
                msg.vini_hmac,
                msg.pseudo_out,
                msg.pseudo_out_hmac,
            )
        except Exception as e:
            await self.tsx_exc_handler(e)
            raise

    async def tsx_all_in_set(self, msg):
        """
        All inputs set. Defining rsig parameters.

        :param msg:
        :return:
        """
        try:
            return await self.tsx_obj.all_in_set()
        except Exception as e:
            await self.tsx_exc_handler(e)
            raise

    async def tsx_set_output1(self, msg):
        """
        Set destination entry one by one.
        Computes destination stealth address, amount key, range proof + HMAC, out_pk, ecdh_info.

        :param msg
        :return:
        """
        try:
            dst, dst_hmac, rsig_data = msg.dst_entr, msg.dst_entr_hmac, msg.rsig_data
            del (msg)

            return await self.tsx_obj.set_out1(dst, dst_hmac, rsig_data)
        except Exception as e:
            await self.tsx_exc_handler(e)
            raise

    async def tsx_all_out1_set(self, msg=None):
        """
        All outputs were set in this phase. Computes additional public keys (if needed), tx.extra and
        transaction prefix hash.
        Adds additional public keys to the tx.extra

        :return: tx.extra, tx_prefix_hash
        """
        try:
            return await self.tsx_obj.all_out1_set()

        except misc.TrezorTxPrefixHashNotMatchingError as e:
            await self.tsx_exc_handler(e)

            from monero_glue.messages.Failure import Failure
            from monero_glue.trezor.wrapper import exc2str

            return Failure(code=10, message=exc2str(e))

        except Exception as e:
            await self.tsx_exc_handler(e)
            raise

    async def tsx_sign_input(self, msg):
        """
        Generates a signature for one input.

        :return:
        """
        try:
            return await self.tsx_obj.sign_input(
                msg.src_entr,
                msg.vini,
                msg.vini_hmac,
                msg.pseudo_out,
                msg.pseudo_out_hmac,
                msg.pseudo_out_alpha,
                msg.spend_key,
            )
        except Exception as e:
            await self.tsx_exc_handler(e)
            raise

    async def tsx_sign_final(self, msg=None):
        """
        Final message.
        Offloading tx related data, encrypted.

        :return:
        """
        try:
            return await self.tsx_obj.final_msg()
        except Exception as e:
            await self.tsx_exc_handler(e)
            raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import logging

import requests
from monero_serialize import xmrserialize, xmrrpc
from requests.auth import HTTPDigestAuth

logger = logging.getLogger(__name__)


class AutoModeler(object):
    def __init__(self):
        pass

    def to_model(self, obj):
        if isinstance(obj, xmrrpc.IModel):
            return obj
        if isinstance(obj, int):
            return xmrrpc.IntegerModel(obj)


class DaemonRpc(object):
    """
    Daemon RPC helper
    """

    def __init__(self, port=28081, creds=None):
        self.port = port
        self.creds = creds
        self.url = None
        self.base_url = None
        self.set_addr("127.0.0.1:%s" % port)

    def set_addr(self, addr):
        self.base_url = "http://%s" % addr
        self.url = "%s/json_rpc" % self.base_url

    def set_creds(self, creds):
        if creds is None or (isinstance(creds, (list, tuple)) and len(creds) == 2):
            self.creds = creds
        elif isinstance(creds, str):
            self.creds = creds.split(":", 1)
        else:
            raise ValueError("Unknown creds type")

    def request(self, method, params=None):
        """
        Request wrapper
        {"jsonrpc":"2.0","id":"0","method":"get_address", "params":{"account_index": 0, "address_index": [0,1,2,3,4,5]}
        :param method:
        :param params:
        :return:
        """
        auth = HTTPDigestAuth(self.creds[0], self.creds[1]) if self.creds else None
        js = {"jsonrpc": "2.0", "id": "0", "method": method}
        if params:
            js["params"] = params

        resp = requests.post(self.url, json=js, auth=auth)
        resp.raise_for_status()
        return resp.json()

    def request_bin(self, method, data=None):
        auth = HTTPDigestAuth(self.creds[0], self.creds[1]) if self.creds else None
        resp = requests.post("%s/%s" % (self.base_url, method), data=data, auth=auth)
        return resp.content

    async def dump_bin(self, msg):
        writer = xmrserialize.MemoryReaderWriter()
        ar = xmrrpc.Archive(writer, True)

        await ar.root()
        await ar.section(msg)
        return writer.get_buffer()

    async def load_bin(self, data):
        reader = xmrserialize.MemoryReaderWriter(bytearray(data))
        ar = xmrrpc.Archive(reader, False)

        msg = {}
        await ar.root()
        await ar.section(msg)
        return msg

    async def get_version(self):
        return self.request("get_version")["result"]

    async def get_info(self):
        return self.request("get_info")["result"]

    async def get_blocks_by_height(self, heights):
        req = {
            "heights": xmrrpc.ArrayModel(
                heights, xmrrpc.SerializeType.UINT64 | xmrrpc.SerializeType.ARRAY_FLAG
            )
        }
        req_bin = await self.dump_bin(req)

        resp_bin = self.request_bin("getblocks_by_height.bin", req_bin)
        resp = await self.load_bin(resp_bin)
        return resp



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .TronContract import TronContract

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class TronSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 252

    def __init__(
        self,
        address_n: List[int] = None,
        ref_block_bytes: bytes = None,
        ref_block_hash: bytes = None,
        expiration: int = None,
        data: str = None,
        contract: TronContract = None,
        timestamp: int = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.ref_block_bytes = ref_block_bytes
        self.ref_block_hash = ref_block_hash
        self.expiration = expiration
        self.data = data
        self.contract = contract
        self.timestamp = timestamp

    @classmethod
    def get_fields(cls):
        return {
            1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            2: ('ref_block_bytes', p.BytesType, 0),
            3: ('ref_block_hash', p.BytesType, 0),
            4: ('expiration', p.UVarintType, 0),
            5: ('data', p.UnicodeType, 0),
            6: ('contract', TronContract, 0),
            7: ('timestamp', p.UVarintType, 0),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import os
import binascii
import collections
import aiounittest
import pkg_resources
from monero_serialize.xmrtypes import RctType

from monero_glue.messages import MoneroRctKeyPublic
from monero_glue.xmr import crypto, monero, ring_ct, common, mlsag2
from monero_glue.hwtoken import misc
from monero_serialize import xmrserialize, xmrtypes

from monero_glue.xmr.enc import chacha_poly
from monero_glue.xmr.sub.addr import get_change_addr_idx
from monero_glue.xmr.sub.seed import SeedDerivation


class BaseAgentTest(aiounittest.AsyncTestCase):

    def get_trezor_mnemonics(self):
        return [
            'permit universe parent weapon amused modify essay borrow tobacco budget '
            'walnut lunch consider gallery ride amazing frog forget treat market '
            'chapter velvet useless topple',

            'permit permit parent weapon amused modify essay borrow tobacco budget '
            'walnut lunch consider gallery ride amazing frog forget treat market '
            'chapter velvet useless topple',

            'permit permit permit weapon amused modify essay borrow tobacco budget '
            'walnut lunch consider gallery ride amazing frog forget treat market '
            'chapter velvet useless topple',
        ]

    def get_trezor_creds(self, idx):
        sd = SeedDerivation.from_mnemonics(self.get_trezor_mnemonics()[idx])
        return sd.creds(monero.NetworkTypes.TESTNET)

    def get_creds(self):
        """
        Wallet credentials
        :return:
        """
        return monero.AccountCreds.new_wallet(
            priv_view_key=crypto.b16_to_scalar(
                b"4ce88c168e0f5f8d6524f712d5f8d7d83233b1e7a2a60b5aba5206cc0ea2bc08"
            ),
            priv_spend_key=crypto.b16_to_scalar(
                b"f2644a3dd97d43e87887e74d1691d52baa0614206ad1b0c239ff4aa3b501750a"
            ),
            network_type=monero.NetworkTypes.TESTNET,
        )

    def get_creds_01(self):
        """
        Wallet 02 credentials
        :return:
        """
        return monero.AccountCreds.new_wallet(
            priv_view_key=crypto.b16_to_scalar(
                b"42ba20adb337e5eca797565be11c9adb0a8bef8c830bccc2df712535d3b8f608"
            ),
            priv_spend_key=crypto.b16_to_scalar(
                b"b0ef6bd527b9b23b9ceef70dc8b4cd1ee83ca14541964e764ad23f5151204f0f"
            ),
            network_type=monero.NetworkTypes.TESTNET,
        )

    def get_creds_02(self):
        """
        Wallet 01 credentials
        :return:
        """
        return monero.AccountCreds.new_wallet(
            priv_view_key=crypto.b16_to_scalar(
                b"9e7aba8ae9ee134e5d5464d9145a4db26793d7411af7d06f20e755cb2a5ad50f"
            ),
            priv_spend_key=crypto.b16_to_scalar(
                b"283d8bab1aeaee8f8b5aed982fc894c67d3e03db9006e488321c053f5183310d"
            ),
            network_type=monero.NetworkTypes.TESTNET,
        )

    def get_trezor_tsx_tests(self):
        # return [('tsx_t_uns_%02d.txt' % i) for i in [1,4,8,14,16,17]]
        return [('tsx_t_uns_%02d.txt' % i) for i in range(1, 18)] \
               + [('tsx_t_uns_%02d_sub.txt' % i) for i in [8, 12]] \
               + [('tsx_t_uns_%02d_sub_add.txt' % i) for i in [8, 12]]

    def get_data_file(self, fl):
        return pkg_resources.resource_string(
            __name__, os.path.join("data", fl)
        )

    async def verify_ki_export(self, res, exp):
        """
        Verifies key image export
        :param res:
        :param exp:
        :return:
        """
        self.assertTrue(len(res) > 0)
        for idx, kie in enumerate(res):
            td = exp.tds[idx]
            ki = crypto.decodepoint(kie[0])
            pkey = crypto.decodepoint(td.m_tx.vout[td.m_internal_output_index].target.key)
            sig = [[crypto.decodeint(kie[1][0]), crypto.decodeint(kie[1][1])]]
            self.assertTrue(ring_ct.check_ring_singature(kie[0], ki, [pkey], sig))

    def get_expected_payment_id(self, fl):
        """
        Expected payment id type, data for file name
        :param fl:
        :return:
        """
        if fl is None:
            return None

        bname = os.path.basename(fl)
        if bname in ['tsx_t_uns_12.txt', 'tsx_t_uns_13.txt', 'tsx_t_uns_14.txt', 'tsx_t_uns_15.txt']:
            return 1, binascii.unhexlify('deadc0dedeadc0de')
        elif bname in ['tsx_t_uns_16.txt', ]:
            return 0, binascii.unhexlify('e8ce72659fbde7cc7c44871ca7784bba24b57323e66f7384ab06f2b8eea40649')
        elif bname in ['tsx_uns01.txt', 'tsx_uns02.txt', 'tsx_uns03.txt']:
            return 1, binascii.unhexlify('deadc0dedeadc0d1')
        elif bname in ['tsx_pending01.txt', ]:
            return 1, binascii.unhexlify('2f3413399032aeea')
        else:
            return None

    async def tx_sign_test(self, agent, con_data, creds, all_creds, fl, sign_tx=False):
        """
        General transaction signer / tester
        :param agent:
        :param con_data:
        :param creds:
        :param all_creds:
        :param fl:
        :param sign_tx ca:
        :return:
        """
        if not sign_tx:
            txes = await agent.sign_unsigned_tx(con_data)
        else:
            txes = await agent.sign_tx(con_data)

        await self.verify(txes[0], agent.last_transaction_data(), creds=creds)
        await self.receive(txes[0], all_creds, agent.last_transaction_data(), self.get_expected_payment_id(fl))

    async def verify(self, tx, con_data=None, creds=None):
        """
        Transaction verification
        :param tx:
        :param con_data:
        :param creds:
        :return:
        """

        # Unserialize the transaction
        tx_obj = xmrtypes.Transaction()
        reader = xmrserialize.MemoryReaderWriter(bytearray(tx))
        ar1 = xmrserialize.Archive(reader, False)

        await ar1.message(tx_obj, msg_type=xmrtypes.Transaction)
        extras = await monero.parse_extra_fields(tx_obj.extra)
        monero.expand_transaction(tx_obj)

        tx_pub = crypto.decodepoint(monero.find_tx_extra_field_by_type(
            extras, xmrtypes.TxExtraPubKey
        ).pub_key)

        additional_pub_keys = monero.find_tx_extra_field_by_type(
            extras, xmrtypes.TxExtraAdditionalPubKeys
        )
        additional_pub_keys = [crypto.decodepoint(x) for x in additional_pub_keys.data] if additional_pub_keys is not None else None

        # Verify range proofs
        out_idx = 0
        is_bp = tx_obj.rct_signatures.type in [RctType.SimpleBulletproof, RctType.FullBulletproof]
        if not is_bp:
            for idx, rsig in enumerate(tx_obj.rct_signatures.p.rangeSigs):
                out_pk = tx_obj.rct_signatures.outPk[idx]
                C = crypto.decodepoint(out_pk.mask)
                rsig = tx_obj.rct_signatures.p.rangeSigs[idx]
                res = ring_ct.ver_range(C, rsig, use_bulletproof=is_bp)
                self.assertTrue(res)

        else:
            for idx, rsig in enumerate(tx_obj.rct_signatures.p.bulletproofs):
                rsig_num_outs = min(len(tx_obj.rct_signatures.outPk), 1 << (len(rsig.L) - 6))
                outs = tx_obj.rct_signatures.outPk[out_idx : out_idx + rsig_num_outs]
                rsig.V = [crypto.encodepoint(ring_ct.bp_comm_to_v(crypto.decodepoint(xx.mask))) for xx in outs]
                res = ring_ct.ver_range(None, rsig, use_bulletproof=is_bp)
                self.assertTrue(res)

        # Prefix hash
        prefix_hash = await monero.get_transaction_prefix_hash(tx_obj)
        is_simple = len(tx_obj.vin) > 1 or is_bp

        self.assertEqual(prefix_hash, con_data.tx_prefix_hash)
        tx_obj.rct_signatures.message = prefix_hash

        # MLSAG hash
        mlsag_hash = await monero.get_pre_mlsag_hash(tx_obj.rct_signatures)

        # Decrypt transaction key
        tx_key = misc.compute_tx_key(creds.spend_key_private, prefix_hash, salt=con_data.enc_salt1, rand_mult=con_data.enc_salt2)[0]
        key_buff = chacha_poly.decrypt_pack(tx_key, con_data.enc_keys)

        tx_priv_keys = [crypto.decodeint(x) for x in common.chunk(key_buff, 32) if x]
        tx_priv = tx_priv_keys[0]
        tx_additional_priv = tx_priv_keys[1:]

        # Verify mlsag signature
        monero.recode_msg(tx_obj.rct_signatures.p.MGs, encode=False)
        for idx in range(len(tx_obj.vin)):
            if is_simple:
                mix_ring = [MoneroRctKeyPublic(dest=x[1].dest, commitment=x[1].mask) for x in con_data.tx_data.sources[idx].outputs]
                if is_bp:
                    pseudo_out = crypto.decodepoint(bytes(tx_obj.rct_signatures.p.pseudoOuts[idx]))
                else:
                    pseudo_out = crypto.decodepoint(bytes(tx_obj.rct_signatures.pseudoOuts[idx]))
                self.assertTrue(mlsag2.ver_rct_mg_simple(
                    mlsag_hash, tx_obj.rct_signatures.p.MGs[idx], mix_ring, pseudo_out
                ))

            else:
                txn_fee_key = crypto.scalarmult_h(tx_obj.rct_signatures.txnFee)
                mix_ring = [[MoneroRctKeyPublic(dest=x[1].dest, commitment=x[1].mask)] for x in con_data.tx_data.sources[idx].outputs]
                self.assertTrue(mlsag2.ver_rct_mg(
                    tx_obj.rct_signatures.p.MGs[idx], mix_ring, tx_obj.rct_signatures.outPk, txn_fee_key, mlsag_hash
                ))

    async def receive(self, tx, all_creds, con_data=None, exp_payment_id=None):
        """
        Test transaction receive with known view/spend keys of destinations.
        :param tx:
        :param all_creds:
        :param con_data:
        :param exp_payment_id:
        :return:
        """
        # Unserialize the transaction
        tx_obj = xmrtypes.Transaction()
        reader = xmrserialize.MemoryReaderWriter(bytearray(tx))
        ar1 = xmrserialize.Archive(reader, False)

        await ar1.message(tx_obj, msg_type=xmrtypes.Transaction)
        extras = await monero.parse_extra_fields(tx_obj.extra)
        tx_pub = monero.find_tx_extra_field_by_type(
            extras, xmrtypes.TxExtraPubKey
        ).pub_key
        additional_pub_keys = monero.find_tx_extra_field_by_type(
            extras, xmrtypes.TxExtraAdditionalPubKeys
        )
        num_outs = len(tx_obj.vout)
        num_received = 0

        # Try to receive tsx outputs with each account.
        tx_money_got_in_outs = collections.defaultdict(lambda: 0)
        outs = []

        change_idx = get_change_addr_idx(con_data.tsx_data.outputs, con_data.tsx_data.change_dts)

        for idx, creds in enumerate(all_creds):
            wallet_subs = {}
            for account in range(0, 5):
                monero.compute_subaddresses(creds, account, range(25), wallet_subs)

            derivation = crypto.generate_key_derivation(
                crypto.decodepoint(tx_pub), creds.view_key_private
            )
            additional_derivations = []
            if additional_pub_keys and additional_pub_keys.data:
                for x in additional_pub_keys.data:
                    additional_derivations.append(
                        crypto.generate_key_derivation(
                            crypto.decodepoint(x), creds.view_key_private
                        )
                    )

            for ti, to in enumerate(tx_obj.vout):
                tx_scan_info = monero.check_acc_out_precomp(
                    to, wallet_subs, derivation, additional_derivations, ti
                )
                if not tx_scan_info.received:
                    continue

                num_received += 1
                tx_scan_info = monero.scan_output(
                    creds, tx_obj, ti, tx_scan_info, tx_money_got_in_outs, outs, False
                )

                # Check spending private key correctness
                self.assertTrue(
                    crypto.point_eq(
                        crypto.decodepoint(tx_obj.rct_signatures.outPk[ti].mask),
                        crypto.gen_c(tx_scan_info.mask, tx_scan_info.amount),
                    )
                )

                self.assertTrue(
                    crypto.point_eq(
                        crypto.decodepoint(tx_obj.vout[ti].target.key),
                        crypto.scalarmult_base(tx_scan_info.in_ephemeral),
                    )
                )

                if exp_payment_id is not None:
                    payment_id = None
                    # Not checking payment id for change transaction
                    if exp_payment_id[0] == 1 and change_idx is not None and ti == change_idx:
                        continue

                    payment_id_type = None
                    extra_nonce = monero.find_tx_extra_field_by_type(extras, xmrtypes.TxExtraNonce)
                    if extra_nonce and monero.has_encrypted_payment_id(extra_nonce.nonce):
                        payment_id_type = 1
                        payment_id = monero.get_encrypted_payment_id_from_tx_extra_nonce(extra_nonce.nonce)
                        payment_id = monero.encrypt_payment_id(payment_id, crypto.decodepoint(tx_pub), creds.view_key_private)

                    elif extra_nonce and monero.has_payment_id(extra_nonce.nonce):
                        payment_id_type = 0
                        payment_id = monero.get_payment_id_from_tx_extra_nonce(extra_nonce.nonce)

                    self.assertEqual(payment_id_type, exp_payment_id[0])
                    self.assertEqual(payment_id, exp_payment_id[1])

        # All outputs have to be successfully received
        self.assertEqual(num_outs, num_received)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

from Crypto.Cipher import AES


def encrypt(key, plaintext, iv=None):
    """
    Uses AES-CBC for encryption
    :param key:
    :param plaintext:
    :param iv:
    :return: ciphertext
    """

    cipher = AES.new(key, AES.MODE_ECB, iv=iv)
    return cipher.encrypt(plaintext)


def decrypt(key, ciphertext, iv=None):
    """
    AES-CBC decryption
    :param key:
    :param ciphertext:
    :param iv:
    :return:
    """
    cipher = AES.new(key, AES.MODE_CBC, iv=iv)
    return cipher.decrypt(ciphertext)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


from monero_serialize import xmrserialize


class TMessage(object):
    def __init__(self, *args, **kwargs):
        self.status = 0

    def __repr__(self):
        dct = (
            xmrserialize.slot_obj_dict(self)
            if hasattr(self, "__slots__")
            else self.__dict__
        )
        return "<%s: %s>" % (self.__class__.__name__, dct)


class TError(TMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status = 1
        self.reason = kwargs.pop("reason", None)
        self.exc = kwargs.pop("exc", None)
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])

    def __repr__(self):
        return "TError(status=%r, reason=%r, exc=%r)" % (
            self.status,
            self.reason,
            self.exc,
        )


class TTxHashNotMatchingError(TError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TResponse(TMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = args
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import hashlib
import hmac
import struct

from Crypto.Protocol.KDF import PBKDF2
from Crypto.Random import get_random_bytes
from Crypto.Random import random as rand

from monero_glue.xmr.core.backend import ed25519ietf
from monero_glue.xmr.core.backend.ed25519 import expmod
from monero_glue.xmr.core.backend.ed25519_2 import inv
from monero_glue.xmr.core.ec_base import *
from monero_serialize import xmrserialize


def cmp(a, b):
    return (a > b) - (a < b)


def memcpy(dst, dst_off, src, src_off, len):
    for i in range(len):
        dst[dst_off + i] = src[src_off + i]
    return dst


_decodeint = ed25519.decodeint
_encodeint = ed25519.encodeint
_encodepoint = ed25519_2.encodepoint
_decodepoint = ed25519_2.decodepoint


class EdScalar(object):
    def __init__(self, v=None, offset=0):
        self.v = 0
        self.init(v, offset)

    def init(self, src=None, offset=0):
        if src is None:
            self.v = 0
        elif isinstance(src, int):
            self.v = src % l
        elif isinstance(src, EdScalar):
            self.v = src.v
        else:
            self.v = _decodeint(src[offset:]) % l
        return self

    def _assert_scalar(self, other):
        if not isinstance(other, EdScalar):
            raise ValueError("operand is not EdScalar")

    def __repr__(self):
        return "EdScalar(%s)" % binascii.hexlify(_encodeint(self.v))

    def __cmp__(self, other):
        if isinstance(other, int):
            return cmp(self.v, other)
        if not isinstance(other, EdScalar):
            raise ValueError("Neither EdScalar nor integer")
        return cmp(self.v, other.v)

    def __eq__(self, other):
        self._assert_scalar(other)
        return self.v == other.v

    def __bytes__(self):
        return _encodeint(self.v)

    def modinv(self):
        self.v = pow(self.v, l - 2, l)
        return self

    def __neg__(self):
        return EdScalar(-1 * self.v)

    def __add__(self, other):
        self._assert_scalar(other)
        return EdScalar(self.v + other.v)

    def __sub__(self, other):
        return EdScalar(self.v - other.v)

    def __mul__(self, other):
        return EdScalar(self.v * other.v)

    @classmethod
    def ensure_scalar(cls, x):
        if isinstance(x, EdScalar):
            return x
        return EdScalar(x)


class EdPoint(object):
    def __init__(self, v=None, offset=0):
        self.v = ed25519_2.ident
        self.init(v, offset)

    def init(self, src=None, offset=0):
        if src is None:
            self.v = ed25519_2.ident
        elif isinstance(src, EdPoint):
            self.v = src.v
        elif isinstance(src, tuple):
            self.v = src
        else:
            self.v = _decodepoint(src[offset:])
        return self

    def __repr__(self):
        return "EdPoint(%r)" % binascii.hexlify(_encodepoint(self.v))

    def __getitem__(self, item):
        return self.v[item]

    def __eq__(self, other):
        if isinstance(other, EdPoint):
            return ed25519ietf.point_equal(self.v, other.v)
        elif isinstance(other, tuple):
            return ed25519ietf.point_equal(self.v, other)
        else:
            ValueError("Neither EdPoint nor quadruple")

    def __bytes__(self):
        return _encodepoint(self.v)

    def check(self):
        if not ed25519_2.isoncurve(self.v):
            raise ValueError("P is not on ed25519 curve")

    @staticmethod
    def invert_v(v):
        return -1 * v[0] % q, v[1], v[2], -1 * v[3] % q

    def invert(self):
        self.v = EdPoint.invert_v(self.v)
        return self

    def _assert_point(self, other):
        if not isinstance(other, EdPoint):
            raise ValueError("operand is not EdPoint")

    def __add__(self, other):
        self._assert_point(other)
        return EdPoint(ed25519_2.edwards_add(self.v, other.v))

    def __neg__(self):
        return EdPoint(self).invert()

    def __sub__(self, other):
        self._assert_point(other)
        return EdPoint(ed25519_2.edwards_add(self.v, EdPoint.invert_v(other.v)))

    def __mul__(self, other):
        return EdPoint(ed25519_2.scalarmult(self.v, other.v))


BASE = EdPoint(ed25519_2.B)


def new_point():
    return EdPoint()


def new_scalar():
    return EdScalar()


def random_bytes(by):
    """
    Generates X random bytes, returns byte-string
    :param by:
    :return:
    """
    return get_random_bytes(by)


def get_keccak():
    """
    Simple keccak 256
    :return:
    """
    return keccak2.Keccak256()


def keccak_hash(inp):
    """
    Hashesh input in one call
    :return:
    """
    ctx = get_keccak()
    ctx.update(inp)
    return ctx.digest()


def keccak_hash_into(r, inp):
    b = keccak_hash(inp)
    return memcpy(r, 0, b, 0, 32)


def keccak_2hash(inp):
    """
    Keccak double hashing
    :param inp:
    :return:
    """
    return keccak_hash(keccak_hash(inp))


def get_hmac(key, msg=None):
    """
    Returns HMAC object (uses Keccak256)
    :param key:
    :param msg:
    :return:
    """
    return hmac.new(key, msg=msg, digestmod=get_keccak)


def compute_hmac(key, msg=None):
    """
    Computes and returns HMAC of the msg using Keccak256
    :param key:
    :param msg:
    :return:
    """
    h = hmac.new(key, msg=msg, digestmod=get_keccak)
    return h.digest()


def pbkdf2(inp, salt, length=32, count=1000, prf=None):
    """
    PBKDF2 with default PRF as SHA256
    HMAC-KECCAK-256 was used before but trezor-crypto does not have pbkdf2 with keccak256.
    :param inp:
    :param salt:
    :param length:
    :param count:
    :param prf:
    :return:
    """

    if prf is None:
        prf = lambda p, s: hmac.new(p, msg=s, digestmod=hashlib.sha256).digest()
    return PBKDF2(inp, salt, length, count, prf)


#
# Basic point enc/dec
#


def _offset(x, offset=0):
    if offset == 0:
        return x
    return x[offset:]


def decodeint(x, offset=0):
    return EdScalar(_offset(x, offset))


def decodeint_into(r, x, offset=0):
    return r.init(_offset(x, offset))


def decodeint_into_noreduce(r, x, offset=0):
    r.v = _decodeint(_offset(x, offset))
    return r


def encodeint(x):
    return bytes(x)


def encodeint_into(b, x, offset=0):
    r = bytes(x)
    return memcpy(b, offset, r, 0, 32)


def check_ed25519point(P):
    P.check()


def encodepoint(P):
    return bytes(P)


def encodepoint_into(b, P, offset=0):
    r = bytes(P)
    return memcpy(b, offset, r, 0, 32)


def conv_xy_to_ext(P):
    """
    Converts x,y representation to extended form
    (x % q, y % q, 1, x*y % q)
    :param P:
    :return:
    """
    return P[0] % q, P[1] % q, 1, (P[0] * P[1]) % q


def conv_ext_to_xy(P):
    """
    Converts extended representation to x,y
    Accepts also projective representation.
    :param P:
    :return:
    """
    x, y, z = P[0], P[1], P[2]
    zi = inv(z)
    x = (x * zi) % q
    y = (y * zi) % q
    return x, y


def conv_xy_to_precomp(P):
    """
    Transform x,y representation to precomputed form
    :param P:
    :return:
    """
    x, y = P
    return y - x, y + x, 2 * ed25519.d * x * y


def conv_ext_to_proj(P):
    """
    Extended coordinates to projective
    :param P:
    :return:
    """
    return P[0], P[1], P[2]


def conv_p1p1_to_ext(P):
    """
    p1p1 representation to extended
    :param P:
    :return:
    """
    x0, y0, z0, t0 = P
    return x0 * t0 % q, y0 * z0 % q, z0 * t0 % q, x0 * y0 % q


def invert_ext(P):
    """
    Inverts extended point coordinate
    :param P:
    :return:
    """
    return P.invert()


def point_eq(P, Q):
    P.check() and Q.check()
    return P == Q


def decodepoint(b, offset=0):
    return EdPoint(_offset(b, offset))


def decodepoint_into(r, b, offset=0):
    return r.init(_offset(b, offset))


#
# Zmod(2^255 - 19) operations, fe (field element)
# Not constant time! PoC only.
#


def fe_1():
    return 1


def fe_mod(a):
    return a % q


def fe_add(a, b):
    return (a + b) % q


def fe_sub(a, b):
    return (a - b) % q


def fe_sq(a):
    return (a * a) % q


def fe_mul(a, b):
    return (a * b) % q


def fe_expmod(b, e):
    return ed25519.expmod(b, e, q)


def fe_divpowm1(u, v):
    """
    uv^3(uv^7)^((q-5)/8)
    :param u:
    :param v:
    :return:
    """
    uv3 = ((u * v) ^ 3) % q
    uv7 = ((u * v) ^ 7) % q

    return uv3 * fe_expmod(uv7, ((q - 5) / 8))


def fe_isnegative(x):
    return x < 0


def fe_isnonzero(x):
    return x != 0


#
# Zmod(order), scalar values field
#


def sc_0():
    return EdScalar(0)


def sc_0_into(r):
    return r.init(0)


def sc_init(x):
    if x >= (1 << 64):
        raise ValueError("Initialization works up to 64-bit only")
    return EdScalar(x)


def sc_init_into(r, x):
    if x >= (1 << 64):
        raise ValueError("Initialization works up to 64-bit only")
    return r.init(x)


def sc_get64(x):
    return x.v


def sc_check(key):
    """
    sc_check is not relevant for long-integer scalar representation.

    :param key:
    :return:
    """
    if key.v % l == 0:
        return -1
    return 0 if key.v == sc_reduce32(key.v) else -1


def check_sc(key):
    """
    throws exception on invalid key
    :param key:
    :return:
    """
    if sc_check(key) != 0:
        raise ValueError("Invalid scalar value")


def sc_reduce32(data):
    return data % l


def sc_add(aa, bb):
    return aa + bb


def sc_add_into(r, aa, bb):
    return r.init(aa + bb)


def sc_sub(aa, bb):
    return aa - bb


def sc_sub_into(r, aa, bb):
    return r.init(aa - bb)


def sc_isnonzero(c):
    return c != ZERO


def sc_eq(a, b):
    return a == b


def sc_mul(a, b):
    return a * b


def sc_mul_into(r, a, b):
    return r.init(a * b)


def sc_mulsub(aa, bb, cc):
    """
    (cc - aa * bb) % l
    """
    return cc - aa * bb


def sc_mulsub_into(r, aa, bb, cc):
    """
    (cc - aa * bb) % l
    """
    return r.init(cc - aa * bb)


def sc_muladd(aa, bb, cc):
    return cc + aa * bb


def sc_muladd_into(r, aa, bb, cc):
    return r.init(cc + aa * bb)


def sc_inv(aa):
    return EdScalar(aa).modinv()


def sc_inv_into(r, x):
    return r.init(x).modinv()


def random_scalar():
    return EdScalar(rand.getrandbits(64 * 8) % l)


def random_scalar_into(r):
    return r.init(random_scalar())


def extended_gcd(aa, bb):
    lastremainder, remainder = abs(aa), abs(bb)
    x, lastx, y, lasty = 0, 1, 1, 0
    while remainder:
        lastremainder, (quotient, remainder) = (
            remainder,
            divmod(lastremainder, remainder),
        )
        x, lastx = lastx - quotient * x, x
        y, lasty = lasty - quotient * y, y
    return lastremainder, lastx * (-1 if aa < 0 else 1), lasty * (-1 if bb < 0 else 1)


def modinv(a, m):
    g, x, y = extended_gcd(a, m)
    if g != 1:
        raise ValueError
    return x % m


def mul_inverse_egcd(x, n, s=1, t=0, N=0):
    return (
        n < 2 and t % N or mul_inverse_egcd(n, x % n, t, s - x // n * t, N or n),
        -1,
    )[n < 1]


#
# GE - ed25519 group
#


def scalarmult_base(a):
    return EdPoint(ed25519_2.scalarmult_B(a.v))


def scalarmult_base_into(r, a):
    return r.init(ed25519_2.scalarmult_B(a.v))


def scalarmult(P, e):
    return P * e


def scalarmult_into(r, P, e):
    return r.init(P * e)


def point_add(A, B):
    return A + B


def point_add_into(r, A, B):
    return r.init(A + B)


def point_sub(A, B):
    return A - B


def point_sub_into(r, A, B):
    return r.init(A - B)


def point_double(P):
    return EdPoint(P + P)


def point_double_into(r, P):
    return r.init(P + P)


def point_norm(P):
    return P


def point_mul8(P):
    return P * EIGHT


INV_EIGHT = b"\x79\x2f\xdc\xe2\x29\xe5\x06\x61\xd0\xda\x1c\x7d\xb3\x9d\xd3\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06"
INV_EIGHT_SC = EdScalar(decodeint(INV_EIGHT))
ZERO = EdScalar(0)
ONE = EdScalar(1)
TWO = EdScalar(2)
EIGHT = EdScalar(8)


def point_mulinv8(P):
    return P * INV_EIGHT_SC


def point_mul8_into(r, P):
    return r.init(P * INV_EIGHT_SC)


def sc_inv_eight():
    return INV_EIGHT_SC


def ge_double_scalarmult_base_vartime(a, A, b):
    """
    void ge_double_scalarmult_base_vartime(ge_p2 *r, const unsigned char *a, const ge_p3 *A, const unsigned char *b)
    r = a * A + b * B

    :param a:
    :param A:
    :param b:
    :return:
    """
    return point_add(scalarmult(A, a), scalarmult_base(b))


def ge_double_scalarmult_base_vartime2(a, A, b, B):
    """
    void ge_double_scalarmult_base_vartime(ge_p2 *r, const unsigned char *a, const ge_p3 *A, const unsigned char *b)
    r = a * A + b * B

    :param a:
    :param A:
    :param b:
    :param B:
    :return:
    """
    return point_add(scalarmult(A, a), scalarmult(B, b))


def ge_double_scalarmult_precomp_vartime(a, A, b, Bi):
    """
    void ge_double_scalarmult_precomp_vartime(ge_p2 *r, const unsigned char *a, const ge_p3 *A, const unsigned char *b, const ge_dsmp Bi)
    :return:
    """
    return ge_double_scalarmult_precomp_vartime2(a, A, b, Bi)


def ge_double_scalarmult_precomp_vartime2(a, Ai, b, Bi):
    """
    void ge_double_scalarmult_precomp_vartime2(ge_p2 *r, const unsigned char *a, const ge_dsmp Ai, const unsigned char *b, const ge_dsmp Bi)
    :param a:
    :param Ai:
    :param b:
    :param Bi:
    :return:
    """
    return point_add(scalarmult(Ai, a), scalarmult(Bi, b))


def identity(byte_enc=False):
    """
    Identity point
    :return:
    """
    idd = EdPoint()
    return idd if not byte_enc else bytes(idd)


def identity_into(r):
    return r.init(ed25519_2.ident)


def ge_frombytes_vartime_check(point):
    """
    https://www.imperialviolet.org/2013/12/25/elligator.html
    http://elligator.cr.yp.to/
    http://elligator.cr.yp.to/elligator-20130828.pdf

    Basically it takes some bytes of data
    converts to a point on the edwards curve
    if the bytes aren't on the curve
    also does some checking on the numbers
    ex. your secret key has to be at least >= 4294967277
    also it rejects certain curve points, i.e. "if x = 0, sign must be positive"

    sqrt(s) = s^((q+3) / 8) if s^((q+3)/4) == s
            = sqrt(-1) s ^((q+3) / 8) otherwise

    :param key:
    :return:
    """
    x, y = point[:2]
    z = fe_1()
    u = fe_sq(y)
    v = fe_mul(u, d)
    u = fe_sub(u, z)  # u = y^2-1
    v = fe_add(v, z)  # v = dy^2+1

    # x = uv^3(uv^7)^((q-5)/8)

    vxx = fe_sq(x)
    vxx = fe_mul(vxx, v)
    check = fe_sub(vxx, u)  # vx^2-u
    if fe_isnonzero(check):
        check = fe_add(vxx, u)
        if fe_isnegative(check):
            # return -1
            raise ValueError("Point check failed")
    return 0


def ge_frombytes_vartime(point):
    """
    https://www.imperialviolet.org/2013/12/25/elligator.html

    :param key:
    :return:
    """
    ge_frombytes_vartime_check(point)
    return point


def precomp(point):
    """
    Precomputation placeholder
    :param point:
    :return:
    """
    return point


def ge_dsm_precomp(point):
    """
    void ge_dsm_precomp(ge_dsmp r, const ge_p3 *s)
    :param point:
    :return:
    """
    return point


#
# Monero specific
#


def cn_fast_hash(buff):
    """
    Keccak 256, original one (before changes made in SHA3 standard)
    :param buff:
    :return:
    """
    kc2 = keccak2.Keccak256()
    kc2.update(buff)
    return kc2.digest()


def hash_to_scalar(data, length=None):
    """
    H_s(P)
    :param data:
    :param length:
    :return:
    """
    hash = cn_fast_hash(data[:length] if length else data)
    return decodeint(hash)


def hash_to_scalar_into(r, data, length=None):
    return r.init(hash_to_scalar(data, length))


def hash_to_point(buf):
    """
    H_p(buf)

    Code adapted from MiniNero: https://github.com/monero-project/mininero
    https://github.com/monero-project/research-lab/blob/master/whitepaper/ge_fromfe_writeup/ge_fromfe.pdf
    http://archive.is/yfINb
    :param key:
    :return:
    """
    u = _decodeint(cn_fast_hash(buf)) % q
    A = 486662

    w = (2 * u * u + 1) % q
    xp = (w * w - 2 * A * A * u * u) % q

    # like sqrt (w / x) although may have to check signs..
    # so, note that if a squareroot exists, then clearly a square exists..
    rx = ed25519.expmod(w * ed25519.inv(xp), (q + 3) // 8, q)
    # rx is ok.

    x = rx * rx * (w * w - 2 * A * A * u * u) % q

    y = (2 * u * u + 1 - x) % q  # w - x, if y is zero, then x = w

    negative = False
    if y != 0:
        y = (w + x) % q  # checking if you got the negative square root.
        if y != 0:
            negative = True

        else:
            rx = rx * -1 * fe_fffb1 % q
            negative = False
    else:
        # y was 0..
        rx = (rx * -1 * fe_fffb2) % q

    if not negative:
        rx = (rx * u) % q
        z = (-2 * A * u * u) % q
        sign = 0

    else:
        z = -1 * A
        x = x * fe_sqrtm1 % q  # ..
        y = (w - x) % q
        if y != 0:
            rx = rx * fe_fffb3 % q
        else:
            rx = rx * -1 * fe_fffb4 % q
        sign = 1

    # setsign
    if (rx % 2) != sign:
        rx = -(rx) % q

    rz = (z + w) % q
    ry = (z - w) % q
    rx = rx * rz % q
    rt = 1

    # extended representation
    rt = ((rx * ry % q) * inv(rz)) % q

    P = EdPoint((rx, ry, rz, rt))
    P8 = scalarmult(P, EdScalar(8))
    return P8


def hash_to_point_into(r, buf):
    return r.init(hash_to_point(buf))


#
# XMR
#


XMR_H = b"\x8b\x65\x59\x70\x15\x37\x99\xaf\x2a\xea\xdc\x9f\xf1\xad\xd0\xea\x6c\x72\x51\xd5\x41\x54\xcf\xa9\x2c\x17\x3a\x0d\xd3\x9c\x1f\x94"
XMR_H_PT = EdPoint(XMR_H)


def compute_H():
    """
    Returns point H
    8b655970153799af2aeadc9ff1add0ea6c7251d54154cfa92c173a0dd39c1f94
    :return:
    """
    h = cn_fast_hash(encodepoint(scalarmult_base(1)))
    return scalarmult(decodepoint(h), 8)


def xmr_H():
    return EdPoint(XMR_H_PT)


def scalarmult_h(i):
    return XMR_H_PT * EdScalar.ensure_scalar(i)


def add_keys2(a, b, B):
    """
    aG + bB, G is basepoint
    :param a:
    :param b:
    :param B:
    :return:
    """
    return point_add(scalarmult_base(a), scalarmult(B, b))


def add_keys2_into(r, a, b, B):
    return r.init(add_keys2(a, b, B))


def add_keys3(a, A, b, B):
    """
    aA + bB
    :param a:
    :param A:
    :param b:
    :param B:
    :return:
    """
    return point_add(scalarmult(A, a), scalarmult(B, b))


def add_keys3_into(r, a, A, b, B):
    return r.init(add_keys3(a, A, b, B))


def gen_c(a, amount):
    """
    Generates Pedersen commitment
    C = aG + bH

    :param a:
    :param amount:
    :return:
    """
    aG = scalarmult_base(a)
    return point_add(aG, scalarmult_h(amount))


def generate_key_derivation(key1, key2):
    """
    Key derivation: 8*(key2*key1)

    :param key1: public key of receiver Bob (see page 7)
    :param key2: Alice's private
    :return:
    """
    if sc_check(key2) != 0:
        # checks that the secret key is uniform enough...
        raise ValueError("error in sc_check in keyder")
    if ge_frombytes_vartime_check(key1) != 0:
        raise ValueError("didn't pass curve checks in keyder")

    check_ed25519point(key1)
    point2 = scalarmult(key1, key2)
    point3 = point_mul8(
        point2
    )  # This has to do with n==0 mod 8 by dedfinition, c.f. the top paragraph of page 5 of http://cr.yp.to/ecdh/curve25519-20060209.pdf
    # and also c.f. middle of page 8 in same document (Bernstein)
    return point3


def derivation_to_scalar(derivation, output_index):
    """
    H_s(derivation || varint(output_index))
    :param derivation:
    :param output_index:
    :return:
    """
    check_ed25519point(derivation)
    buf2 = encodepoint(derivation) + xmrserialize.dump_uvarint_b(output_index)
    return hash_to_scalar(buf2, len(buf2))


def derive_public_key(derivation, output_index, base):
    """
    H_s(derivation || varint(output_index))G + base

    :param derivation:
    :param output_index:
    :param base:
    :return:
    """
    if ge_frombytes_vartime_check(base) != 0:  # check some conditions on the point
        raise ValueError("derive pub key bad point")
    check_ed25519point(base)

    scalar = derivation_to_scalar(derivation, output_index)
    point2 = scalarmult_base(scalar)
    point4 = point_add(base, point2)
    return point4


def derive_secret_key(derivation, output_index, base):
    """
    base + H_s(derivation || varint(output_index))
    :param derivation:
    :param output_index:
    :param base:
    :return:
    """
    if sc_check(base) != 0:
        raise ValueError("cs_check in derive_secret_key")
    scalar = derivation_to_scalar(derivation, output_index)
    return sc_add(base, scalar)


def get_subaddress_secret_key(secret_key, major=0, minor=0):
    """
    Builds subaddress secret key from the subaddress index
    Hs(SubAddr || a || index_major || index_minor)

    :param secret_key:
    :param index:
    :param major:
    :param minor:
    :return:
    """
    prefix = b"SubAddr"
    buffer = bytearray(len(prefix) + 1 + 32 + 4 + 4)
    struct.pack_into(
        "<7sb32sLL", buffer, 0, prefix, 0, encodeint(secret_key), major, minor
    )
    return hash_to_scalar(buffer)


#
# Backend config
#


class PyECBackend(ECBackendBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def has_crypto_into_functions(self):
        return True

    def is_fast(self):
        return False


BACKEND_OBJ = None


def get_backend():
    global BACKEND_OBJ
    if BACKEND_OBJ is None:
        BACKEND_OBJ = PyECBackend()
    return BACKEND_OBJ



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import binascii

from monero_glue.misc.bip import bip32, bip39, bip39_deriv
from monero_glue.xmr import crypto, monero
from monero_glue.xmr.core import mnemonic
from monero_glue.xmr.sub.xmr_net import NetworkTypes


DEFAULT_BIP44_PATH = "m/44'/128'/0'/0/0"
DEFAULT_SLIP0010_PATH = "m/44'/128'/0'"


class SeedDerivation(object):
    def __init__(self):
        self.mnemonics = None
        self.mnemonics_as_idx = False
        self.master_seed = None
        self.pre_hash = None
        self.monero_master = None
        self.electrum_words = None
        self.is_slip0010 = False
        self.path = None

        self.spend_sec = None
        self.spend_pub = None
        self.view_sec = None
        self.view_pub = None

    def set_seed(self, seed, path=None, slip0010=False):
        """
        Sets master secret for BIP44 derivation
        :param seed:
        :param path:
        :param slip0010:
        :return:
        """
        self.master_seed = seed
        self.is_slip0010 = slip0010
        if path is None:
            self.path = DEFAULT_BIP44_PATH if not slip0010 else DEFAULT_SLIP0010_PATH
        else:
            self.path = path
        wl = bip32.Wallet.from_master_secret(seed, use_ed25519=slip0010)

        # Generate private keys based on the gen mechanism. Bip44 path + Monero backward compatible
        data = wl.get_child_for_path(self.path)
        self.pre_hash = binascii.unhexlify(data.private_key.get_key())

        if slip0010:
            self.monero_master = crypto.encodeint(crypto.decodeint(self.pre_hash))

        else:
            # Ledger way = words -> bip39 pbkdf -> master seed -> bip32 normal with
            #  "Bitcoin seed" seed, get private key node -> cn_fast_hash -> monero master secret
            self.monero_master = crypto.cn_fast_hash(self.pre_hash)

        self.set_monero_seed(self.monero_master)

    def set_monero_seed(self, seed):
        """
        Sets Monero master secret seed.
        :param seed:
        :return:
        """
        # to_hash is initial seed in the Monero sense, recoverable from this seed
        self.monero_master = seed
        self.electrum_words = " ".join(mnemonic.mn_encode(self.monero_master, True))

        keys = monero.generate_monero_keys(self.monero_master)
        self.spend_sec, self.spend_pub, self.view_sec, self.view_pub = keys

    def creds(self, network_type=NetworkTypes.MAINNET):
        return monero.AccountCreds.new_wallet(
            priv_view_key=self.view_sec,
            priv_spend_key=self.spend_sec,
            network_type=network_type,
        )

    @staticmethod
    def clean_input(inp):
        cleaned = []
        if isinstance(inp, (str, bytes)):
            cleaned = [inp]

        else:
            for w in inp:
                cleaned += w.split(" ")

        cleaned = [x.strip().lower() for x in cleaned]
        return cleaned

    @classmethod
    def from_mnemonics(cls, mnemonics, as_index=False, *args, **kwargs):
        mnems = SeedDerivation.clean_input(mnemonics)

        if as_index:
            indices = [bip39.english_words.index(x) for x in mnems]
            seed = bip32.Wallet.indices_to_bytes(indices)

        else:
            seed = bip39_deriv.mnemonics_to_seed(" ".join(mnems))

        r = cls()
        r.mnemonics = mnems
        r.mnemonics_as_idx = as_index
        r.set_seed(seed, *args, **kwargs)
        return r

    @classmethod
    def from_master_seed(cls, seed, *args, **kwargs):
        r = cls()
        r.set_seed(seed, *args, **kwargs)
        return r

    @classmethod
    def from_monero_seed(cls, seed, *args, **kwargs):
        r = cls()
        r.set_monero_seed(seed)
        return r

    @classmethod
    def from_monero_mnemonics(cls, mnemonics_words, *args, **kwargs):
        mnems = SeedDerivation.clean_input(mnemonics_words)
        seed = mnemonic.mn_decode(mnems)

        r = cls()
        r.set_monero_seed(binascii.unhexlify(seed))
        return r



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import binascii
import json
import os
import unittest

import aiounittest
import pkg_resources

from monero_glue.messages import MoneroRctKeyPublic
from monero_glue.xmr import crypto, mlsag2, monero, ring_ct
from monero_serialize import xmrserialize, xmrtypes


class MoneroTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(MoneroTest, self).__init__(*args, **kwargs)

    def test_wallet_addr(self):
        addr = monero.encode_addr(
            monero.net_version(),
            binascii.unhexlify(
                b"3bec484c5d7f0246af520aab550452b5b6013733feabebd681c4a60d457b7fc1"
            ),
            binascii.unhexlify(
                b"2d5918e31d3c003da3c778592c07b398ad6f961a67082a75fd49394d51e69bbe"
            ),
        )
        self.assertEqual(
            addr,
            b"43tpGG9PKbwCpjRvNLn1jwXPpnacw2uVUcszAtgmDiVcZK4VgHwjJT9BJz1WGF9eMxSYASp8yNMkuLjeQfWqJn3CNWdWfzV",
        )

        w = monero.AccountCreds.new_wallet(
            crypto.b16_to_scalar(
                b"4ce88c168e0f5f8d6524f712d5f8d7d83233b1e7a2a60b5aba5206cc0ea2bc08"
            ),
            crypto.b16_to_scalar(
                b"f2644a3dd97d43e87887e74d1691d52baa0614206ad1b0c239ff4aa3b501750a"
            ),
            network_type=monero.NetworkTypes.TESTNET,
        )
        self.assertEqual(
            w.address,
            b"9vacMKaj8JJV6MnwDzh2oNVdwTLJfTDyNRiB6NzV9TT7fqvzLivH2dB8Tv7VYR3ncn8vCb3KdNMJzQWrPAF1otYJ9cPKpkr",
        )

    def test_derive_subaddress_public_key(self):
        out_key = crypto.decodepoint(
            bytes(
                [
                    0xf4,
                    0xef,
                    0xc2,
                    0x9d,
                    0xa4,
                    0xcc,
                    0xd6,
                    0xbc,
                    0x6e,
                    0x81,
                    0xf5,
                    0x2a,
                    0x6f,
                    0x47,
                    0xb2,
                    0x95,
                    0x29,
                    0x66,
                    0x44,
                    0x2a,
                    0x7e,
                    0xfb,
                    0x49,
                    0x90,
                    0x1c,
                    0xce,
                    0x06,
                    0xa7,
                    0xa3,
                    0xbe,
                    0xf3,
                    0xe5,
                ]
            )
        )
        deriv = crypto.decodepoint(
            bytes(
                [
                    0x25,
                    0x9e,
                    0xf2,
                    0xab,
                    0xa8,
                    0xfe,
                    0xb4,
                    0x73,
                    0xcf,
                    0x39,
                    0x05,
                    0x8a,
                    0x0f,
                    0xe3,
                    0x0b,
                    0x9f,
                    0xf6,
                    0xd2,
                    0x45,
                    0xb4,
                    0x2b,
                    0x68,
                    0x26,
                    0x68,
                    0x7e,
                    0xbd,
                    0x6b,
                    0x63,
                    0x12,
                    0x8a,
                    0xff,
                    0x64,
                ]
            )
        )
        res = crypto.encodepoint(monero.derive_subaddress_public_key(out_key, deriv, 5))
        self.assertEqual(
            res,
            bytes(
                [
                    0x5a,
                    0x10,
                    0xcc,
                    0xa9,
                    0x00,
                    0xee,
                    0x47,
                    0xa7,
                    0xf4,
                    0x12,
                    0xcd,
                    0x66,
                    0x1b,
                    0x29,
                    0xf5,
                    0xab,
                    0x35,
                    0x6d,
                    0x6a,
                    0x19,
                    0x51,
                    0x88,
                    0x45,
                    0x93,
                    0xbb,
                    0x17,
                    0x0b,
                    0x5e,
                    0xc8,
                    0xb6,
                    0xf2,
                    0xe8,
                ]
            ),
        )

    def test_get_subaddress_secret_key(self):
        a = crypto.b16_to_scalar(
            b"4ce88c168e0f5f8d6524f712d5f8d7d83233b1e7a2a60b5aba5206cc0ea2bc08"
        )
        m = monero.get_subaddress_secret_key(secret_key=a, major=0, minor=1)
        self.assertEqual(
            crypto.encodeint(m),
            bytes(
                [
                    0xb6,
                    0xff,
                    0x4d,
                    0x68,
                    0x9b,
                    0x95,
                    0xe3,
                    0x31,
                    0x0e,
                    0xfb,
                    0xf6,
                    0x83,
                    0x85,
                    0x0c,
                    0x07,
                    0x5b,
                    0xcd,
                    0xe4,
                    0x63,
                    0x61,
                    0x92,
                    0x30,
                    0x54,
                    0xe4,
                    0x2e,
                    0xf3,
                    0x00,
                    0x16,
                    0xb2,
                    0x87,
                    0xff,
                    0x0c,
                ]
            ),
        )

    def test_public_spend(self):
        derivation = bytes(
            [
                0xe7,
                0x20,
                0xa0,
                0x9f,
                0x2e,
                0x3a,
                0x0b,
                0xbf,
                0x4e,
                0x4b,
                0xa7,
                0xad,
                0x93,
                0x65,
                0x3b,
                0xb2,
                0x96,
                0x88,
                0x55,
                0x10,
                0x12,
                0x1f,
                0x80,
                0x6a,
                0xcb,
                0x2a,
                0x5f,
                0x91,
                0x68,
                0xfa,
                0xfa,
                0x01,
            ]
        )
        base = bytes(
            [
                0x7d,
                0x99,
                0x6b,
                0x0f,
                0x2d,
                0xb6,
                0xdb,
                0xb5,
                0xf2,
                0xa0,
                0x86,
                0x21,
                0x1f,
                0x23,
                0x99,
                0xa4,
                0xa7,
                0x47,
                0x9b,
                0x2c,
                0x91,
                0x1a,
                0xf3,
                0x07,
                0xfd,
                0xc3,
                0xf7,
                0xf6,
                0x1a,
                0x88,
                0xcb,
                0x0e,
            ]
        )
        pkey_ex = bytes(
            [
                0x08,
                0x46,
                0xca,
                0xe7,
                0x40,
                0x50,
                0x77,
                0xb6,
                0xb7,
                0x80,
                0x0f,
                0x0b,
                0x93,
                0x2c,
                0x10,
                0xa1,
                0x86,
                0x44,
                0x83,
                0x70,
                0xb6,
                0xdb,
                0x31,
                0x8f,
                0x8c,
                0x9e,
                0x13,
                0xf7,
                0x81,
                0xda,
                0xb5,
                0x46,
            ]
        )
        pkey_comp = crypto.derive_public_key(
            crypto.decodepoint(derivation), 0, crypto.decodepoint(base)
        )
        self.assertEqual(pkey_ex, crypto.encodepoint(pkey_comp))

    async def test_node_transaction(self):
        tx_j = pkg_resources.resource_string(
            __name__, os.path.join("data", "tsx_01.json")
        )
        tx_c = pkg_resources.resource_string(
            __name__, os.path.join("data", "tsx_01_plain.txt")
        )
        tx_u_c = pkg_resources.resource_string(
            __name__, os.path.join("data", "tsx_01_uns.txt")
        )
        tx_js = json.loads(tx_j.decode("utf8"))

        reader = xmrserialize.MemoryReaderWriter(bytearray(binascii.unhexlify(tx_c)))
        ar = xmrserialize.Archive(reader, False)
        tx = xmrtypes.Transaction()
        await ar.message(tx)

        reader = xmrserialize.MemoryReaderWriter(bytearray(binascii.unhexlify(tx_u_c)))
        ar = xmrserialize.Archive(reader, False)
        uns = xmrtypes.UnsignedTxSet()
        await ar.message(uns)

        # Test message hash computation
        tx_prefix_hash = await monero.get_transaction_prefix_hash(tx)
        message = binascii.unhexlify(tx_js["tx_prefix_hash"])
        self.assertEqual(tx_prefix_hash, message)

        # RingCT, range sigs, hash
        rv = tx.rct_signatures
        rv.message = message
        rv.mixRing = self.mixring(tx_js)
        digest = await monero.get_pre_mlsag_hash(rv)
        full_message = binascii.unhexlify(tx_js["pre_mlsag_hash"])
        self.assertEqual(digest, full_message)

        # Recompute missing data
        monero.expand_transaction(tx)

        # Unmask ECDH data, check range proofs
        for i in range(len(tx_js["amount_keys"])):
            ecdh = monero.copy_ecdh(rv.ecdhInfo[i])
            monero.recode_ecdh(ecdh, encode=False)

            ecdh = ring_ct.ecdh_decode(
                ecdh, derivation=binascii.unhexlify(tx_js["amount_keys"][i])
            )
            self.assertEqual(crypto.sc_get64(ecdh.amount), tx_js["outamounts"][i])
            self.assertTrue(
                crypto.sc_eq(
                    ecdh.mask,
                    crypto.decodeint(binascii.unhexlify(tx_js["outSk"][i])[32:]),
                )
            )

            C = crypto.decodepoint(rv.outPk[i].mask)
            rsig = rv.p.rangeSigs[i]

            res = ring_ct.ver_range(C, rsig)
            self.assertTrue(res)

            res = ring_ct.ver_range(
                crypto.point_add(C, crypto.scalarmult_base(crypto.sc_init(3))), rsig
            )
            self.assertFalse(res)

        is_simple = len(tx.vin) > 1
        monero.recode_rct(rv, encode=False)

        if is_simple:
            for index in range(len(rv.p.MGs)):
                pseudo_out = crypto.decodepoint(
                    binascii.unhexlify(
                        tx_js["tx"]["rct_signatures"]["pseudoOuts"][index]
                    )
                )
                r = mlsag2.ver_rct_mg_simple(
                    full_message, rv.p.MGs[index], rv.mixRing[index], pseudo_out
                )
                self.assertTrue(r)

                r = mlsag2.ver_rct_mg_simple(
                    full_message, rv.p.MGs[index], rv.mixRing[index - 1], pseudo_out
                )
                self.assertFalse(r)

        else:
            txn_fee_key = crypto.scalarmult_h(rv.txnFee)
            r = mlsag2.ver_rct_mg(
                rv.p.MGs[0], rv.mixRing, rv.outPk, txn_fee_key, digest
            )
            self.assertTrue(r)

            r = mlsag2.ver_rct_mg(
                rv.p.MGs[0],
                rv.mixRing,
                rv.outPk,
                crypto.scalarmult_h(rv.txnFee - 100),
                digest,
            )
            self.assertFalse(r)

    def mixring(self, js):
        mxr = []
        mx = js["mixRing"]
        for i in range(len(mx)):
            mxr.append([])
            for j in range(len(mx[i])):
                dt = binascii.unhexlify(mx[i][j])
                mxr[i].append(MoneroRctKeyPublic(dest=dt[:32], commitment=dt[32:]))
        return mxr


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



from monero_glue.protocol import tsx_sign
from monero_glue.protocol import tsx_sign_state
from monero_glue.protocol import tsx_sign_state_holder
__all__ = ['tsx_sign', 'tsx_sign_state', 'tsx_sign_state_holder']



# https://tools.ietf.org/html/draft-josefsson-eddsa-ed25519-00#section-4
# Copyright (c) <YEAR>, <OWNER>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

import hashlib

from . import ed25519


def sha512(s):
    return hashlib.sha512(s).digest()


# Base field Z_p
p = 2 ** 255 - 19


def modp_inv(x):
    return pow(x, p - 2, p)


# Curve constant
d = -121665 * modp_inv(121666) % p


# Group order
q = 2 ** 252 + 27742317777372353535851937790883648493

# Points are represented as tuples (X, Y, Z, T) of extended coordinates,
# with x = X/Z, y = Y/Z, x*y = T/Z


def point_add(P, Q):
    A = (P[1] - P[0]) * (Q[1] - Q[0]) % p
    B = (P[1] + P[0]) * (Q[1] + Q[0]) % p
    C = 2 * P[3] * Q[3] * d % p
    D = 2 * P[2] * Q[2] % p
    E = B - A
    F = D - C
    G = D + C
    H = B + A
    return (E * F, G * H, F * G, E * H)


# Computes Q = s * Q
def point_mul(s, P):
    Q = (0, 1, 1, 0)  # Neutral element
    while s > 0:
        # Is there any bit-set predicate?
        if s & 1:
            Q = point_add(Q, P)
        P = point_add(P, P)
        s >>= 1
    return Q


def point_equal(P, Q):
    # x1 / z1 == x2 / z2  <==>  x1 * z2 == x2 * z1
    if (P[0] * Q[2] - Q[0] * P[2]) % p != 0:
        return False
    if (P[1] * Q[2] - Q[1] * P[2]) % p != 0:
        return False
    return True


# Square root of -1
modp_sqrt_m1 = pow(2, (p - 1) // 4, p)


# Compute corresponding x coordinate, with low bit corresponding to sign,
# or return None on failure
def recover_x(y, sign):
    x2 = (y * y - 1) * modp_inv(d * y * y + 1)
    if x2 == 0:
        if sign:
            return None
        else:
            return 0

    # Compute square root of x2
    x = pow(x2, (p + 3) // 8, p)
    if (x * x - x2) % p != 0:
        x = x * modp_sqrt_m1 % p
    if (x * x - x2) % p != 0:
        return None

    if (x & 1) != sign:
        x = p - x
    return x


# Base point
g_y = 4 * modp_inv(5) % p
g_x = recover_x(g_y, 0)
G = (g_x, g_y, 1, g_x * g_y % p)


def basepoint():
    return G


def point_compress(P):
    zinv = modp_inv(P[2])
    x = P[0] * zinv % p
    y = P[1] * zinv % p
    return ed25519.encodeint(y | ((x & 1) << 255))


def point_decompress(s):
    # if len(s) != 32:
    # raise Exception("Invalid input length for decompression")
    # y = int.from_bytes(s, "little")
    y = ed25519.decodeint(s)
    sign = y >> 255
    y &= (1 << 255) - 1

    x = recover_x(y, sign)
    if x is None:
        return None
    else:
        return x, y, 1, x * y % p



import argparse
import logging
import re

import coloredlogs

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.INFO, use_chroot=False)


def avg(iterable):
    return sum(iterable) / float(len(iterable))


class MemReader(object):
    def __init__(self, data):
        self.data = data
        self.cline = -1

    def full_line(self):
        return self.data[self.cline]

    def next_section(self):
        offset = 0
        res = self.next_entry()
        if res is None:
            return None
        while not res[1]:
            offset += 1
            res = self.next_entry()
            if res is None:
                return None
        return res[0], res[1], offset

    def next(self, to_section=False):
        if to_section:
            return self.next_section()
        else:
            return self.next_entry()

    def next_entry(self):
        for i in range(self.cline + 1, len(self.data)):
            self.cline = i
            line = self.data[i]

            ma = re.match(r"^.*A:\s*(\d+).*", line)
            if ma is None:
                continue

            mfree = int(ma.group(1))
            is_section = "####" in line

            return mfree, is_section

        return None, None


class MemCompare(object):
    def __init__(self):
        self.args = None

    def read_files(self, files):
        file_data = []

        for idx, file in enumerate(files):
            with open(file) as fh:
                file_data.append(fh.readlines())

        if len(file_data) != 2:
            raise ValueError("Exactly 2 files are required")

        ma = MemReader(file_data[0])
        mb = MemReader(file_data[1])

        m_a_sec = []
        m_b_sec = []
        m_a_all = []
        m_b_all = []
        m_diffs_sec = []
        m_diffs_all = []

        while True:
            ar = ma.next()
            br = mb.next()
            a_off = 0
            b_off = 0

            # End
            if ar is None or br is None:
                if ar is None and br is not None:
                    logger.info("Source A ended before B")
                elif ar is not None and br is None:
                    logger.info("Source B ended before A")
                break

            # Align sections
            if ar[1] and not br[1]:
                br = mb.next_section()
                b_off = br[2]
                logger.info("Skipped %s mem lines from B" % b_off)

            elif not ar[1] and br[1]:
                ar = ma.next_section()
                a_off = ar[2]
                logger.info("Skipped %s mem lines from A" % a_off)

            m_a_all.append(ar[0])
            m_b_all.append(br[0])

            if ar[1]:
                m_a_sec.append(ar[0])
            if br[1]:
                m_b_sec.append(br[0])

            m_diffs_all.append(ar[0] - br[0])
            if ar[1] and br[1]:
                m_diffs_sec.append(ar[0] - br[0])

            if self.args.verbose:
                pass

        avg_all_a = avg(m_a_all)
        avg_all_b = avg(m_b_all)
        avg_sec_a = avg(m_a_sec)
        avg_sec_b = avg(m_b_sec)

        print("Diffs sec: %s" % avg(m_diffs_sec))
        print("Diffs all: %s\n" % avg(m_diffs_all))

        print(
            "Avg Sec A: %s  Sec B: %s  diff: %s   ratio:  %s"
            % (
                avg_sec_a,
                avg_sec_b,
                avg_sec_a - avg_sec_b,
                avg_sec_a / float(avg_sec_b),
            )
        )
        print(
            "Avg All A: %s  All B: %s  diff: %s   ratio:  %s"
            % (
                avg_all_a,
                avg_all_b,
                avg_all_a - avg_all_b,
                avg_all_a / float(avg_all_b),
            )
        )

    def main(self):
        parser = argparse.ArgumentParser(
            description="Compares two log files and used memory"
        )
        parser.add_argument(
            "--verbose",
            dest="verbose",
            default=False,
            action="store_const",
            const=True,
            help="Append to the tee file",
        )
        parser.add_argument(
            "files",
            metavar="FILE",
            nargs="*",
            help="files to read, if empty, stdin is used",
        )
        args = parser.parse_args()
        self.args = args

        try:
            self.read_files(args.files)

        except KeyboardInterrupt:
            logger.info("Terminating")


if __name__ == "__main__":
    anz = MemCompare()
    anz.main()



from monero_glue.xmr.sub import addr
from monero_glue.xmr.sub import creds
from monero_glue.xmr.sub import keccak_hasher
from monero_glue.xmr.sub import mlsag_hasher
from monero_glue.xmr.sub import recode
from monero_glue.xmr.sub import recode_ext
from monero_glue.xmr.sub import tsx_helper
from monero_glue.xmr.sub import xmr_net


from monero_glue.xmr import crypto
from monero_glue.xmr.sub.recode import recode_msg
from monero_serialize import xmrtypes


def recode_rangesig(rsig, encode=True, copy=False):
    """
    In - place rsig recoding
    :param rsig:
    :param encode: if true encodes to byte representation, otherwise decodes from byte representation
    :param copy:
    :return:
    """
    recode_int = crypto.encodeint if encode else crypto.decodeint
    recode_point = crypto.encodepoint if encode else crypto.decodepoint
    nrsig = rsig
    if copy:
        nrsig = xmrtypes.RangeSig()
        nrsig.Ci = [None] * 64
        nrsig.asig = xmrtypes.BoroSig()
        nrsig.asig.s0 = [None] * 64
        nrsig.asig.s1 = [None] * 64

    for i in range(len(rsig.Ci)):
        nrsig.Ci[i] = recode_point(rsig.Ci[i])
    for i in range(len(rsig.asig.s0)):
        nrsig.asig.s0[i] = recode_int(rsig.asig.s0[i])
    for i in range(len(rsig.asig.s1)):
        nrsig.asig.s1[i] = recode_int(rsig.asig.s1[i])
    nrsig.asig.ee = recode_int(rsig.asig.ee)
    return nrsig


def flatten_rsig(rsig):
    """
    Rsig -> byte array
    :param rsig:
    :return:
    """
    res = b""

    for i in range(len(rsig.asig.s0)):
        res += bytes(rsig.asig.s0[i])
    for i in range(len(rsig.asig.s1)):
        res += bytes(rsig.asig.s1[i])
    res += bytes(rsig.asig.ee)
    for i in range(len(rsig.Ci)):
        res += bytes(rsig.Ci[i])
    return res


def inflate_rsig(buff, rsig=None):
    """
    Rsig binary repr -> byte encoded repr
    :param rsig:
    :return:
    """
    if rsig is None:
        rsig = xmrtypes.RangeSig()
        rsig.Ci = [None] * 64
        rsig.asig = xmrtypes.BoroSig()
        rsig.asig.s0 = [None] * 64
        rsig.asig.s1 = [None] * 64

    for i in range(64):
        rsig.asig.s0[i], buff = buff[:32], buff[32:]
    for i in range(64):
        rsig.asig.s1[i], buff = buff[:32], buff[32:]
    rsig.asig.ee, buff = buff[:32], buff[32:]
    for i in range(64):
        rsig.Ci[i], buff = buff[:32], buff[32:]
    return rsig


def recode_rct(rv, encode=True):
    """
    Recodes RCT MGs signatures from raw forms to bytearrays so it works with serialization
    :param rv:
    :param encode: if true encodes to byte representation, otherwise decodes from byte representation
    :return:
    """
    rv.p.MGs = recode_msg(rv.p.MGs, encode=encode)
    return rv



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import os

try:
    from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305

    ChaCha20Poly1305(b"\x00" * 32)  # test if openssl provides ChaCha20Poly1305

except:
    from chacha20poly1305 import ChaCha20Poly1305


def encrypt(key, plaintext, associated_data=None):
    """
    Uses ChaCha20Poly1305 for encryption
    :param key:
    :param plaintext:
    :param associated_data:
    :return: iv, ciphertext, tag
    """
    cipher = ChaCha20Poly1305(key)
    nonce = os.urandom(12)
    ciphertext = cipher.encrypt(bytes(nonce), bytes(plaintext), associated_data)
    return nonce, ciphertext, b""


def decrypt(key, iv, ciphertext, associated_data=None):
    """
    ChaCha20Poly1305 decryption
    :param key:
    :param iv:
    :param ciphertext:
    :param associated_data:
    :return:
    """
    cipher = ChaCha20Poly1305(key)
    plaintext = cipher.decrypt(iv, ciphertext, associated_data)
    return plaintext


def encrypt_pack(key, plaintext, associated_data=None):
    return b"".join(encrypt(key, plaintext, associated_data))


def decrypt_pack(key, ciphertext):
    return decrypt(key, ciphertext[:12], ciphertext[12:], None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

from monero_glue import protobuf
from monero_glue.compat import gc
from monero_glue.messages import (
    MoneroAccountPublicAddress,
    MoneroMultisigKLRki,
    MoneroOutputEntry,
    MoneroRctKeyPublic,
    MoneroTransactionData,
    MoneroTransactionDestinationEntry,
    MoneroTransactionSourceEntry,
)
from monero_glue.xmr import crypto
from monero_serialize import xmrserialize, xmrtypes


class TrezorError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])


class TrezorSecurityError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorInvalidStateError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorTxPrefixHashNotMatchingError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorChangeAddressError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class StdObj(object):
    def __init__(self, **kwargs):
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])


def compute_tx_key(spend_key_private, tx_prefix_hash, salt=None, rand_mult=None):
    if not salt:
        salt = crypto.random_bytes(32)

    if not rand_mult:
        rand_mult_num = crypto.random_scalar()
        rand_mult = crypto.encodeint(rand_mult_num)
    else:
        rand_mult_num = crypto.decodeint(rand_mult)

    rand_inp = crypto.sc_add(spend_key_private, rand_mult_num)
    passwd = crypto.keccak_2hash(crypto.encodeint(rand_inp) + tx_prefix_hash)
    tx_key = crypto.compute_hmac(salt, passwd)
    # tx_key = crypto.pbkdf2(passwd, salt, count=100)
    return tx_key, salt, rand_mult


def translate_monero_dest_entry_pb(dst_entry: xmrtypes.TxDestinationEntry):
    d = MoneroTransactionDestinationEntry(
        amount=dst_entry.amount,
        is_subaddress=dst_entry.is_subaddress,
        addr=MoneroAccountPublicAddress(
            spend_public_key=dst_entry.addr.m_spend_public_key,
            view_public_key=dst_entry.addr.m_view_public_key,
        ),
    )
    return d


def translate_monero_src_entry_pb(src_entry: xmrtypes.TxSourceEntry):
    d = MoneroTransactionSourceEntry()
    d.outputs = [
        MoneroOutputEntry(
            idx=x[0], key=MoneroRctKeyPublic(dest=x[1].dest, commitment=x[1].mask)
        )
        for x in src_entry.outputs
    ]
    d.real_output = src_entry.real_output
    d.real_out_tx_key = src_entry.real_out_tx_key
    d.real_out_additional_tx_keys = list(src_entry.real_out_additional_tx_keys)
    d.real_output_in_tx_index = src_entry.real_output_in_tx_index
    d.amount = src_entry.amount
    d.rct = src_entry.rct
    d.mask = src_entry.mask
    if src_entry.multisig_kLRki:
        s = src_entry.multisig_kLRki
        src_entry.multisig_kLRki = MoneroMultisigKLRki(K=s.K, L=s.L, R=s.R, ki=s.ki)
    return d


async def parse_msg(bts, msg):
    reader = xmrserialize.MemoryReaderWriter(bytearray(bts))
    ar = xmrserialize.Archive(reader, False)
    return await ar.message(msg)


async def parse_pb_msg(bts, msg):
    reader = xmrserialize.MemoryReaderWriter(bytearray(bts))
    return await protobuf.load_message(reader, msg)


async def parse_vini(bts):
    return await parse_msg(bts, xmrtypes.TxinToKey())


async def dump_msg(msg, preallocate=None, msg_type=None, prefix=None):
    writer = xmrserialize.MemoryReaderWriter(preallocate=preallocate)
    if prefix:
        writer.write(prefix)
    ar = xmrserialize.Archive(writer, True)
    await ar.message(msg, msg_type=msg_type)
    return writer.get_buffer()


async def dump_pb_msg(msg):
    writer = xmrserialize.MemoryReaderWriter()
    await protobuf.dump_message(writer, msg)
    return bytes(writer.get_buffer())


async def dump_msg_gc(msg, preallocate=None, msg_type=None, del_msg=False):
    b = await dump_msg(msg, preallocate=preallocate, msg_type=msg_type)
    if del_msg:
        del msg

    import gc

    gc.collect()
    return b


def dump_rsig_bp(rsig):
    from monero_glue.compat.utils import memcpy

    if len(rsig.L) > 127:
        raise ValueError("Too large")

    # Manual serialization as the generic purpose misc.dump_msg_gc
    # is more memory intensive which is not desired in the range proof section.

    # BP: V, A, S, T1, T2, taux, mu, L, R, a, b, t
    # Commitment vector V is not serialized
    # Vector size under 127 thus varint occupies 1 B
    buff_size = 32 * (9 + 2 * (len(rsig.L))) + 2
    buff = bytearray(buff_size)

    memcpy(buff, 0, rsig.A, 0, 32)
    memcpy(buff, 32, rsig.S, 0, 32)
    memcpy(buff, 32 * 2, rsig.T1, 0, 32)
    memcpy(buff, 32 * 3, rsig.T2, 0, 32)
    memcpy(buff, 32 * 4, rsig.taux, 0, 32)
    memcpy(buff, 32 * 5, rsig.mu, 0, 32)

    buff[32 * 6] = len(rsig.L)
    offset = 32 * 6 + 1

    for x in rsig.L:
        memcpy(buff, offset, x, 0, 32)
        offset += 32

    buff[offset] = len(rsig.R)
    offset += 1

    for x in rsig.R:
        memcpy(buff, offset, x, 0, 32)
        offset += 32

    memcpy(buff, offset, rsig.a, 0, 32)
    offset += 32
    memcpy(buff, offset, rsig.b, 0, 32)
    offset += 32
    memcpy(buff, offset, rsig.t, 0, 32)
    return buff


def dst_entry_to_stdobj(dst):
    if dst is None:
        return None

    addr = StdObj(
        spend_public_key=dst.addr.spend_public_key,
        view_public_key=dst.addr.view_public_key,
    )
    return StdObj(amount=dst.amount, addr=addr, is_subaddress=dst.is_subaddress)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: https://github.com/monero-project/mininero
# Author: Dusan Klinec, ph4r05, 2018

import logging

from monero_glue.compat import gc
from monero_glue.compat.utils import memcpy
from monero_glue.xmr import bulletproof, crypto, mlsag2, monero
from monero_serialize import xmrtypes

logger = logging.getLogger(__name__)
ATOMS = 64


def d2b(n, digits):
    b = [0] * digits
    i = 0
    while n:
        b[i] = n & 1
        i = i + 1
        n >>= 1
    return b


def sum_Ci(Cis):
    """
    Sums points
    :param Cis:
    :return:
    """
    CSum = crypto.identity()
    for i in Cis:
        CSum = crypto.point_add(CSum, i)
    return CSum


def bp_size(outputs):
    M, logM = 1, 0
    while M <= 16 and M < outputs:
        logM += 1
        M = 1 << logM

    return 32 * (21 + outputs + 2 * logM)


async def prove_range_bp(amount, last_mask=None):
    mask = last_mask if last_mask is not None else crypto.random_scalar()
    bp_proof = await prove_range_bp_batch([amount], [mask])

    C = crypto.decodepoint(bp_proof.V[0])
    C = crypto.point_mul8(C)

    gc.collect()

    # Return as struct as the hash(BP_struct) != hash(BP_serialized)
    # as the original hashing does not take vector lengths into account which are dynamic
    # in the serialization scheme (and thus extraneous)
    return C, mask, bp_proof


async def prove_range_bp_batch(amounts, masks):
    from monero_glue.xmr import bulletproof as bp

    bpi = bp.BulletProofBuilder()
    bp_proof = bpi.prove_batch([crypto.sc_init(a) for a in amounts], masks)
    del (bpi, bp)
    gc.collect()

    return bp_proof


async def verify_bp(bp_proof, amounts=None, masks=None):
    from monero_glue.xmr import bulletproof as bp

    if amounts:
        bp_proof.V = []
        for i in range(len(amounts)):
            C = crypto.gen_c(masks[i], amounts[i])
            crypto.scalarmult_into(C, C, crypto.sc_inv_eight())
            bp_proof.V.append(crypto.encodepoint(C))

    bpi = bp.BulletProofBuilder()
    res = bpi.verify(bp_proof)
    gc.collect()

    # Return as struct as the hash(BP_struct) != hash(BP_serialized)
    # as the original hashing does not take vector lengths into account which are dynamic
    # in the serialization scheme (and thus extraneous)
    return res


def bp_comm_to_v(C):
    """
    Commitment to proof.V
    In mainnet it holds C = 8V
    :param C:
    :return:
    """
    return crypto.point_mulinv8(C)


def prove_range_chunked(amount, last_mask=None):
    a = crypto.sc_init(0)
    si = crypto.sc_init(0)
    c = crypto.sc_init(0)
    ee = crypto.sc_init(0)
    tmp_ai = crypto.sc_init(0)
    tmp_alpha = crypto.sc_init(0)

    C_acc = crypto.identity()
    C_h = crypto.xmr_H()
    C_tmp = crypto.identity()
    L = crypto.identity()
    Zero = crypto.identity()
    kck = crypto.get_keccak()

    ai = bytearray(32 * 64)
    alphai = bytearray(32 * 64)
    buff = bytearray(32)

    Cis = bytearray(32 * 64)
    s0s = bytearray(32 * 64)
    s1s = bytearray(32 * 64)
    ee_bin = bytearray(32)

    for ii in range(64):
        crypto.random_scalar_into(tmp_ai)
        if last_mask is not None and ii == 63:
            crypto.sc_sub_into(tmp_ai, last_mask, a)

        crypto.sc_add_into(a, a, tmp_ai)
        crypto.random_scalar_into(tmp_alpha)

        crypto.scalarmult_base_into(L, tmp_alpha)
        crypto.scalarmult_base_into(C_tmp, tmp_ai)

        # C_tmp += &Zero if BB(ii) == 0 else &C_h
        crypto.point_add_into(C_tmp, C_tmp, Zero if ((amount >> ii) & 1) == 0 else C_h)
        crypto.point_add_into(C_acc, C_acc, C_tmp)

        # Set Ci[ii] to sigs
        crypto.encodepoint_into(Cis, C_tmp, ii << 5)
        crypto.encodeint_into(ai, tmp_ai, ii << 5)
        crypto.encodeint_into(alphai, tmp_alpha, ii << 5)

        if ((amount >> ii) & 1) == 0:
            crypto.random_scalar_into(si)
            crypto.encodepoint_into(buff, L)
            crypto.hash_to_scalar_into(c, buff)

            crypto.point_sub_into(C_tmp, C_tmp, C_h)
            crypto.add_keys2_into(L, si, c, C_tmp)

            crypto.encodeint_into(s1s, si, ii << 5)

        crypto.encodepoint_into(buff, L)
        kck.update(buff)

        crypto.point_double_into(C_h, C_h)

    # Compute ee
    tmp_ee = kck.digest()
    crypto.decodeint_into(ee, tmp_ee)
    del (tmp_ee, kck)

    C_h = crypto.xmr_H()
    gc.collect()

    # Second pass, s0, s1
    for ii in range(64):
        crypto.decodeint_into(tmp_alpha, alphai, ii << 5)
        crypto.decodeint_into(tmp_ai, ai, ii << 5)

        if ((amount >> ii) & 1) == 0:
            crypto.sc_mulsub_into(si, tmp_ai, ee, tmp_alpha)
            crypto.encodeint_into(s0s, si, ii << 5)

        else:
            crypto.random_scalar_into(si)
            crypto.encodeint_into(s0s, si, ii << 5)

            crypto.decodepoint_into(C_tmp, Cis, ii << 5)
            crypto.add_keys2_into(L, si, ee, C_tmp)
            crypto.encodepoint_into(buff, L)
            crypto.hash_to_scalar_into(c, buff)

            crypto.sc_mulsub_into(si, tmp_ai, c, tmp_alpha)
            crypto.encodeint_into(s1s, si, ii << 5)

        crypto.point_double_into(C_h, C_h)

    crypto.encodeint_into(ee_bin, ee)

    del (ai, alphai, buff, tmp_ai, tmp_alpha, si, c, ee, C_tmp, C_h, L, Zero)
    gc.collect()

    return C_acc, a, [s0s, s1s, ee_bin, Cis]


def prove_range(
    amount,
    last_mask=None,
    mem_opt=True,
    backend_impl=False,
    decode=False,
    byte_enc=False,
    rsig=None,
):
    """
    Range proof generator.
    In order to minimize the memory consumption and CPU usage during transaction generation the returned values
    are returned encoded.

    :param amount:
    :param last_mask:
    :param mem_opt: memory optimized
    :param backend_impl: backend implementation, if available
    :param decode: decodes output
    :param byte_enc: decodes output
    :param rsig: buffer for rsig
    :return:
    """
    if backend_impl and crypto.get_backend().has_rangeproof_borromean():
        if byte_enc and decode:
            raise ValueError("Conflicting options byte_enc, decode")

        C, a, R = crypto.prove_range(amount, last_mask)[:3]  # backend returns encoded
        if not byte_enc:
            R = monero.inflate_rsig(R)
        if decode:
            R = monero.inflate_rsig(R)
            R = monero.recode_rangesig(R, encode=False)

        return C, a, R

    ret = None
    if not mem_opt:
        ret = prove_range_orig(amount, last_mask=last_mask)
    else:
        ret = prove_range_mem(amount, last_mask=last_mask)

    # Encoding
    C, a, R = ret[:3]
    if byte_enc:
        R = monero.recode_rangesig(R, encode=True)
        R = monero.flatten_rsig(R)
    elif not decode:
        R = monero.recode_rangesig(R, encode=True)

    return C, a, R


def prove_range_orig(amount, last_mask=None):
    """
    Gives C, and mask such that \sumCi = C
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1

    Ci is a commitment to either 0 or 2^i, i=0,...,63
    thus this proves that "amount" is in [0, 2^ATOMS]
    mask is a such that C = aG + bH, and b = amount
    :param amount:
    :param last_mask: ai[ATOMS-1] will be computed as \sum_{i=0}^{ATOMS-2} a_i - last_mask
    :return: sumCi, mask, RangeSig.
        sumCi is Pedersen commitment on the amount value. sumCi = aG + amount*H
        mask is "a" from the Pedersent commitment above.
    """
    bb = d2b(amount, ATOMS)  # gives binary form of bb in "digits" binary digits
    logger.info("amount, amount in binary %s %s" % (amount, bb))
    ai = [None] * len(bb)
    Ci = [None] * len(bb)
    CiH = [None] * len(bb)  # this is like Ci - 2^i H
    H2 = crypto.gen_Hpow(ATOMS)
    a = crypto.sc_0()
    for i in range(0, ATOMS):
        ai[i] = crypto.random_scalar()
        if last_mask is not None and i == ATOMS - 1:
            ai[i] = crypto.sc_sub(last_mask, a)

        a = crypto.sc_add(
            a, ai[i]
        )  # creating the total mask since you have to pass this to receiver...
        if bb[i] == 0:
            Ci[i] = crypto.scalarmult_base(ai[i])
        if bb[i] == 1:
            Ci[i] = crypto.point_add(crypto.scalarmult_base(ai[i]), H2[i])
        CiH[i] = crypto.point_sub(Ci[i], H2[i])

    A = xmrtypes.BoroSig()
    A.s0, A.s1, A.ee = mlsag2.gen_borromean(ai, Ci, CiH, bb)

    R = xmrtypes.RangeSig()
    R.asig = A
    R.Ci = Ci

    C = sum_Ci(Ci)
    return C, a, R


def prove_range_mem(amount, last_mask=None):
    """
    Memory optimized range proof.

    Gives C, and mask such that \sumCi = C
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1

    Ci is a commitment to either 0 or 2^i, i=0,...,63
    thus this proves that "amount" is in [0, 2^ATOMS]
    mask is a such that C = aG + bH, and b = amount
    :param amount:
    :param last_mask: ai[ATOMS-1] will be computed as \sum_{i=0}^{ATOMS-2} a_i - last_mask
    :return: sumCi, mask, RangeSig.
        sumCi is Pedersen commitment on the amount value. sumCi = aG + amount*H
        mask is "a" from the Pedersent commitment above.
    """
    n = ATOMS
    bb = d2b(amount, n)  # gives binary form of bb in "digits" binary digits
    ai = [None] * len(bb)
    Ci = [None] * len(bb)
    a = crypto.sc_0()

    C = crypto.identity()
    alpha = mlsag2.key_zero_vector(n)
    s1 = mlsag2.key_zero_vector(n)
    c_H = crypto.xmr_H()
    kck = crypto.get_keccak()  # ee computation

    # First pass, generates: ai, alpha, Ci, ee, s1
    for ii in range(n):
        ai[ii] = crypto.random_scalar()
        if last_mask is not None and ii == ATOMS - 1:
            ai[ii] = crypto.sc_sub(last_mask, a)

        a = crypto.sc_add(
            a, ai[ii]
        )  # creating the total mask since you have to pass this to receiver...

        alpha[ii] = crypto.random_scalar()
        L = crypto.scalarmult_base(alpha[ii])

        if bb[ii] == 0:
            Ci[ii] = crypto.scalarmult_base(ai[ii])
        else:
            Ci[ii] = crypto.point_add(crypto.scalarmult_base(ai[ii]), c_H)
        C = crypto.point_add(C, Ci[ii])

        if bb[ii] == 0:
            s1[ii] = crypto.random_scalar()
            c = crypto.hash_to_scalar(crypto.encodepoint(L))
            L = crypto.add_keys2(s1[ii], c, crypto.point_sub(Ci[ii], c_H))
            kck.update(crypto.encodepoint(L))

        else:
            kck.update(crypto.encodepoint(L))

        c_H = crypto.point_double(c_H)

    # Compute ee, memory cleanup
    ee = crypto.decodeint(kck.digest())
    del kck

    # Second phase computes: s0, s1
    c_H = crypto.xmr_H()
    s0 = mlsag2.key_zero_vector(n)

    for jj in range(n):
        if not bb[jj]:
            s0[jj] = crypto.sc_mulsub(ai[jj], ee, alpha[jj])

        else:
            s0[jj] = crypto.random_scalar()
            LL = crypto.add_keys2(s0[jj], ee, Ci[jj])
            cc = crypto.hash_to_scalar(crypto.encodepoint(LL))
            s1[jj] = crypto.sc_mulsub(ai[jj], cc, alpha[jj])
        c_H = crypto.point_double(c_H)

    A = xmrtypes.BoroSig()
    A.s0, A.s1, A.ee = s0, s1, ee

    R = xmrtypes.RangeSig()
    R.asig = A
    R.Ci = Ci

    return C, a, R


def ver_range(C=None, rsig=None, use_bulletproof=False, decode=True):
    """
    Verifies that \sum Ci = C and that each Ci is a commitment to 0 or 2^i
    :param C:
    :param rsig:
    :param use_bulletproof: bulletproof
    :param decode: decodes encoded range proof
    :return:
    """
    n = ATOMS
    CiH = [None] * n
    C_tmp = crypto.identity()
    c_H = crypto.xmr_H()

    if decode and not use_bulletproof:
        rsig = monero.recode_rangesig(rsig, encode=False, copy=True)

    if not use_bulletproof:
        for i in range(0, n):
            CiH[i] = crypto.point_sub(rsig.Ci[i], c_H)
            C_tmp = crypto.point_add(C_tmp, rsig.Ci[i])
            c_H = crypto.point_double(c_H)

        if C is not None and not crypto.point_eq(C_tmp, C):
            return 0

    if use_bulletproof:
        bp = bulletproof.BulletProofBuilder()
        return bp.verify(rsig)

    return mlsag2.ver_borromean(rsig.Ci, CiH, rsig.asig.s0, rsig.asig.s1, rsig.asig.ee)


# Ring-ct MG sigs
# Prove:
#   c.f. http:#eprint.iacr.org/2015/1098 section 4. definition 10.
#   This does the MG sig on the "dest" part of the given key matrix, and
#   the last row is the sum of input commitments from that column - sum output commitments
#   this shows that sum inputs = sum outputs
# Ver:
#   verifies the above sig is created corretly


def ecdh_encode(unmasked, receiver_pk=None, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    :param unmasked:
    :param receiver_pk:
    :param derivation:
    :return:
    """
    rv = xmrtypes.EcdhTuple()
    if derivation is None:
        esk = crypto.random_scalar()
        rv.senderPk = crypto.scalarmult_base(esk)
        derivation = crypto.encodepoint(crypto.scalarmult(receiver_pk, esk))

    sec1 = crypto.hash_to_scalar(derivation)
    sec2 = crypto.hash_to_scalar(crypto.encodeint(sec1))

    rv.mask = crypto.sc_add(unmasked.mask, sec1)
    rv.amount = crypto.sc_add(unmasked.amount, sec2)
    return rv


def ecdh_encode_into(dst, unmasked, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    """
    sec1 = crypto.hash_to_scalar(derivation)
    sec2 = crypto.hash_to_scalar(crypto.encodeint(sec1))

    dst.mask = crypto.sc_add(unmasked.mask, sec1)
    dst.amount = crypto.sc_add(unmasked.amount, sec2)
    return dst


def ecdh_decode(masked, receiver_sk=None, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    """
    rv = xmrtypes.EcdhTuple()

    if derivation is None:
        derivation = crypto.scalarmult(masked.senderPk, receiver_sk)

    sharedSec1 = crypto.hash_to_scalar(derivation)
    sharedSec2 = crypto.hash_to_scalar(crypto.encodeint(sharedSec1))

    rv.mask = crypto.sc_sub(masked.mask, sharedSec1)
    rv.amount = crypto.sc_sub(masked.amount, sharedSec2)
    return rv


#
# RingCT protocol
#


def decode_rct(rv, sk, i):
    """
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1.1
    Uses the attached ecdh info to find the amounts represented by each output commitment
    must know the destination private key to find the correct amount, else will return a random number

    :param rv:
    :param sk:
    :param i:
    :return:
    """
    decodedTuple = ecdh_decode(rv.ecdhInfo[i], sk)
    mask = decodedTuple.mask
    amount = decodedTuple.amount
    C = rv.outPk[i].mask
    H = crypto.xmr_H()
    Ctmp = crypto.point_add(crypto.scalarmult_base(mask), crypto.scalarmult(H, amount))
    if not crypto.point_eq(crypto.point_sub(C, Ctmp), crypto.identity()):
        logger.warning("warning, amount decoded incorrectly, will be unable to spend")
    return amount


#
# Key image import / export
#


def generate_ring_signature(prefix_hash, image, pubs, sec, sec_idx, test=False):
    """
    Generates ring signature with key image.
    void crypto_ops::generate_ring_signature()
    """
    if test:
        t = crypto.scalarmult_base(sec)
        if not crypto.point_eq(t, pubs[sec_idx]):
            raise ValueError("Invalid sec key")

        k_i = monero.generate_key_image(crypto.encodepoint(pubs[sec_idx]), sec)
        if not crypto.point_eq(k_i, image):
            raise ValueError("Key image invalid")
        for k in pubs:
            crypto.ge_frombytes_vartime_check(k)

    image_unp = crypto.ge_frombytes_vartime(image)
    image_pre = crypto.ge_dsm_precomp(image_unp)

    buff_off = len(prefix_hash)
    buff = bytearray(buff_off + 2 * 32 * len(pubs))
    memcpy(buff, 0, prefix_hash, 0, buff_off)
    mvbuff = memoryview(buff)

    sum = crypto.sc_0()
    k = crypto.sc_0()
    sig = []
    for i in range(len(pubs)):
        sig.append([crypto.sc_0(), crypto.sc_0()])  # c, r

    for i in range(len(pubs)):
        if i == sec_idx:
            k = crypto.random_scalar()
            tmp3 = crypto.scalarmult_base(k)
            crypto.encodepoint_into(mvbuff[buff_off : buff_off + 32], tmp3)
            buff_off += 32

            tmp3 = crypto.hash_to_point(crypto.encodepoint(pubs[i]))
            tmp2 = crypto.scalarmult(tmp3, k)
            crypto.encodepoint_into(mvbuff[buff_off : buff_off + 32], tmp2)
            buff_off += 32

        else:
            sig[i] = [crypto.random_scalar(), crypto.random_scalar()]
            tmp3 = crypto.ge_frombytes_vartime(pubs[i])
            tmp2 = crypto.ge_double_scalarmult_base_vartime(sig[i][0], tmp3, sig[i][1])
            crypto.encodepoint_into(mvbuff[buff_off : buff_off + 32], tmp2)
            buff_off += 32

            tmp3 = crypto.hash_to_point(crypto.encodepoint(tmp3))
            tmp2 = crypto.ge_double_scalarmult_precomp_vartime(
                sig[i][1], tmp3, sig[i][0], image_pre
            )
            crypto.encodepoint_into(mvbuff[buff_off : buff_off + 32], tmp2)
            buff_off += 32

            sum = crypto.sc_add(sum, sig[i][0])

    h = crypto.hash_to_scalar(buff)
    sig[sec_idx][0] = crypto.sc_sub(h, sum)
    sig[sec_idx][1] = crypto.sc_mulsub(sig[sec_idx][0], sec, k)
    return sig


def check_ring_singature(prefix_hash, image, pubs, sig):
    """
    Checks ring signature generated with generate_ring_signature
    """
    image_unp = crypto.ge_frombytes_vartime(image)
    image_pre = crypto.ge_dsm_precomp(image_unp)

    buff_off = len(prefix_hash)
    buff = bytearray(buff_off + 2 * 32 * len(pubs))
    memcpy(buff, 0, prefix_hash, 0, buff_off)
    mvbuff = memoryview(buff)

    sum = crypto.sc_0()
    for i in range(len(pubs)):
        if crypto.sc_check(sig[i][0]) != 0 or crypto.sc_check(sig[i][1]) != 0:
            return False

        tmp3 = crypto.ge_frombytes_vartime(pubs[i])
        tmp2 = crypto.ge_double_scalarmult_base_vartime(sig[i][0], tmp3, sig[i][1])
        crypto.encodepoint_into(mvbuff[buff_off : buff_off + 32], tmp2)
        buff_off += 32

        tmp3 = crypto.hash_to_point(crypto.encodepoint(pubs[i]))
        tmp2 = crypto.ge_double_scalarmult_precomp_vartime(
            sig[i][1], tmp3, sig[i][0], image_pre
        )
        crypto.encodepoint_into(mvbuff[buff_off : buff_off + 32], tmp2)
        buff_off += 32

        sum = crypto.sc_add(sum, sig[i][0])

    h = crypto.hash_to_scalar(buff)
    h = crypto.sc_sub(h, sum)
    return crypto.sc_isnonzero(h) == 0


def export_key_image(
    creds,
    subaddresses,
    pkey,
    tx_pub_key,
    additional_tx_pub_keys,
    out_idx,
    test=True,
    verify=True,
):
    """
    Generates key image for the TXO + signature for the key image
    """
    r = monero.generate_key_image_helper(
        creds, subaddresses, pkey, tx_pub_key, additional_tx_pub_keys, out_idx
    )
    xi, ki, recv_derivation = r[:3]

    phash = crypto.encodepoint(ki)
    sig = generate_ring_signature(phash, ki, [pkey], xi, 0, test)

    if verify:
        if check_ring_singature(phash, ki, [pkey], sig) != 1:
            raise ValueError("Signature error")

    return ki, sig
